from functools import partial

from ewokscore.missing_data import is_missing_data
from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread
from silx.gui import qt

from darfix.gui.concatenate_scans import HDF5ConcatenateWindow
from darfix.gui.configuration.level import ConfigurationLevel
from darfix.tasks.hdf5_scans_concatenation import ConcatenateHDF5Scans


class ConcatenateWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=ConcatenateHDF5Scans):
    """
    Widget that concatenates a set of scans.

    Note: 'output_detector_data_path' and 'output_positioners_data_path' are never set by the GUI.
    It simplifies usage. We 'force' users to use the default one.
    """

    name = "HDF5 scans concatenation"
    icon = "icons/concatenate_hdf5.svg"
    want_main_area = True
    want_control_area = False

    _ewoks_inputs_to_hide_from_orange = (
        "input_file",
        "output_file",
        "entries_to_concatenate",
        "detector_data_path",
        "positioners_group_path",
        "output_detector_data_path",
        "output_positioners_data_path",
        "overwrite",
        "guess_output_file",
    )

    def __init__(self):
        super().__init__()
        self._widget = HDF5ConcatenateWindow(parent=self)
        self.mainArea.layout().addWidget(self._widget)

        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self.mainArea.layout().addWidget(self._buttons)

        # load settings
        self._loadSettings()

        # connect signal / slot
        concatenateWidget = self._widget._mainWidget
        self._buttons.button(qt.QDialogButtonBox.Ok).released.connect(self.validate)
        concatenateWidget.sigInputFileChanged.connect(self._inputFileChanged)
        concatenateWidget.sigOutputFileChanged.connect(self._outputFileChanged)
        concatenateWidget.sigOverwriteChanged.connect(self._overwriteOutputChanged)
        concatenateWidget.sigDetectorPathChanged.connect(self._detectorPathChanged)
        concatenateWidget.sigMetadataPathChanged.connect(self._metadataPathChanged)
        concatenateWidget.sigEntriesToConcatenateChanged.connect(
            self._entriesToConcatenateChanged
        )
        concatenateWidget._autoUpdate.toggled.connect(self._autoUpdateChanged)

        self.task_executor.finished.connect(
            self.information,
        )
        self.task_executor.started.connect(
            partial(self.information, "Concatenating scans...")
        )

    def _loadSettings(self):
        upgrade_settings_mode = False
        concatenateWidget = self._widget._mainWidget
        input_file = self.get_task_input_value("input_file")
        if not is_missing_data(input_file):
            concatenateWidget.setInputFile(input_file)
            # if the file is not set, skip settings entries. There will be None but that is fine
            entries_to_concatenate = self.get_task_input_value("entries_to_concatenate")
            if not is_missing_data(entries_to_concatenate):
                concatenateWidget.setSelectedEntries(entries=entries_to_concatenate)

        output_file = self.get_task_input_value("output_file")
        if not is_missing_data(output_file):
            concatenateWidget.setOutputFile(output_file)

        overwrite = self.get_task_input_value("overwrite")
        if not is_missing_data(overwrite):
            concatenateWidget.setOverwrite(overwrite)
        else:
            # update overwrite value because the value from the gui is different from the task one
            self._overwriteOutputChanged(overwrite=concatenateWidget.getOverwrite())

        detector_data_path = self.get_task_input_value("detector_data_path")
        if not is_missing_data(detector_data_path):
            concatenateWidget.setDetectorPath(detector_data_path)
        else:
            # make sure the pattern is registered
            self._detectorPathChanged(concatenateWidget._detectorDataPath.getPattern())

        positioners_group_path = self.get_task_input_value("positioners_group_path")
        if not is_missing_data(positioners_group_path):
            concatenateWidget.setPositionersGroupPath(positioners_group_path)
        else:
            # make sure the pattern is registered
            self._metadataPathChanged(
                concatenateWidget._positionerDataPath.getPattern()
            )

        guess_output_file = self.get_task_input_value("guess_output_file")
        if not is_missing_data(guess_output_file):
            concatenateWidget._autoUpdate.setChecked(guess_output_file)
            if not guess_output_file:
                upgrade_settings_mode = True

        if upgrade_settings_mode:
            # if an advanced option has been set, display all the settings
            concatenateWidget.setConfigurationLevel(ConfigurationLevel.ADVANCED)

    def validate(self):
        self.execute_ewoks_task()
        self.accept()

    def handleNewSignals(self) -> None:
        pass
        # Do not call super().handleNewSignals() to make sure the processing is not triggered

    # ewoks input setter
    def _inputFileChanged(self, file_path: str):
        self.set_default_input("input_file", file_path)

    def _outputFileChanged(self, file_path: str):
        self.set_default_input("output_file", file_path)

    def _overwriteOutputChanged(self, overwrite: bool):
        self.set_default_input("overwrite", overwrite)

    def _detectorPathChanged(self, data_path: str):
        self.set_default_input("detector_data_path", data_path)

    def _metadataPathChanged(self, data_path: str):
        self.set_default_input("positioners_group_path", data_path)

    def _entriesToConcatenateChanged(self):
        self.set_default_input(
            "entries_to_concatenate", self._widget._mainWidget.getSelectedEntries()
        )

    def _autoUpdateChanged(self, activated: bool):
        self.set_default_input("guess_output_file", activated)



import numpy
import pytest

try:
    import scipy
except ImportError:
    scipy = None

from darfix.core import imageRegistration
from darfix.tests import utils


@pytest.fixture
def data():

    return numpy.array(
        [
            [
                [1, 2, 3, 4, 5],
                [2, 2, 3, 4, 5],
                [3, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
            ],
            [
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 3],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
            ],
            [
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [8, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
            ],
        ]
    )


@pytest.fixture
def rstate():
    return numpy.random.RandomState(1000)


@pytest.fixture
def metadata_dict():
    return utils.create_2_motors_metadata(2, 5)


@pytest.fixture
def first_frame(rstate):
    first_frame = numpy.zeros((100, 100))
    first_frame[30:40, 30:40] = rstate.randint(50, 100, size=(10, 10))

    return first_frame


def test_find_shift(data):
    """Tests the shift found"""
    shift = (-1.4, 1.32)
    for img in data:
        # The shift corresponds to the pixel offset relative to the reference image
        offset_image = imageRegistration._opencv_fft_shift(img, shift[1], shift[0])
        computed_shift = imageRegistration.find_shift(img, offset_image, 100)
        numpy.testing.assert_allclose(shift, -computed_shift, rtol=1e-04)


def test_apply_shift(data):
    """Tests the correct apply of the shift"""
    shift = (0, 0)
    for img in data:
        shifted_image = imageRegistration.apply_opencv_shift(img, shift)
        numpy.testing.assert_allclose(img, shifted_image, rtol=1e-04)
    for img in data:
        shifted_image = imageRegistration.apply_opencv_shift(
            img, shift, shift_approach="linear"
        )
        numpy.testing.assert_allclose(img, shifted_image, rtol=1e-04)


def test_improve_shift(data):
    """Tests the shift improvement"""
    h = imageRegistration.improve_linear_shift(
        data, [1, 1], 0.1, 0.1, 1, shift_approach="fft"
    )
    numpy.testing.assert_allclose(h, 0.0)
    h = imageRegistration.improve_linear_shift(
        data, [1, 1], 0.1, 0.1, 1, shift_approach="linear"
    )
    numpy.testing.assert_allclose(h, 0.0)


@pytest.mark.skipif(scipy is None, reason="scipy is missing")
def test_shift_detection10(rstate):
    """Tests the shift detection with tolerance of 3 decimals"""
    first_frame = numpy.zeros((100, 100))
    # Simulating a series of frame with information in the middle.
    first_frame[25:75, 25:75] = rstate.randint(50, 300, size=(50, 50))
    data = [first_frame]
    shift = [1.0, 0]
    for i in range(9):
        data += [
            numpy.fft.ifftn(
                scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
            ).real
        ]
    data = numpy.asanyarray(data, dtype=numpy.int16)
    optimal_shift = imageRegistration.shift_detection(data, 100, shift_approach="fft")

    shift = [-1, 0]

    numpy.testing.assert_allclose(shift, numpy.round(optimal_shift))


@pytest.mark.skipif(scipy is None, reason="scipy is missing")
def test_shift_detection01(rstate):
    """Tests the shift detection with tolerance of 5 decimals"""
    # Create a frame and repeat it shifting it every time
    first_frame = numpy.zeros((100, 100))
    first_frame[25:75, 25:75] = rstate.randint(50, 300, size=(50, 50))
    data = [first_frame]
    shift = [0, 1]
    for i in range(9):
        data += [
            numpy.fft.ifftn(
                scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
            ).real
        ]
    data = numpy.asanyarray(data, dtype=numpy.int16)
    optimal_shift = imageRegistration.shift_detection(data, 100, shift_approach="fft")

    shift = [0, -1]

    numpy.testing.assert_allclose(shift, numpy.round(optimal_shift))


@pytest.mark.skipif(scipy is None, reason="scipy is missing")
def test_shift_detection11(rstate):
    """Tests the shift detection with tolerance of 2 decimals"""
    # Create a frame and repeat it shifting it every time
    first_frame = numpy.zeros((100, 100))
    first_frame[25:75, 25:75] = rstate.randint(50, 300, size=(50, 50))
    data = [first_frame]
    shift = [1, 1]
    for i in range(9):
        data += [
            numpy.fft.ifftn(
                scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
            ).real
        ]
    data = numpy.asanyarray(data, dtype=numpy.int16)

    optimal_shift = imageRegistration.shift_detection(data, 100)

    shift = [-1, -1]

    numpy.testing.assert_allclose(shift, numpy.round(optimal_shift))


@pytest.mark.skipif(scipy is None, reason="scipy is missing")
def test_shift_detection_float(rstate):
    """Tests the shift detection using shifted float with tolerance of 2 decimals"""
    first_frame = numpy.zeros((100, 100))
    # Simulating a series of frame with information in the middle.
    first_frame[25:75, 25:75] = rstate.randint(50, 300, size=(50, 50))
    data = [first_frame]
    shift = [0.5, 0.2]
    for i in range(9):
        data += [
            numpy.fft.ifftn(
                scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
            ).real
        ]
    data = numpy.asanyarray(data, dtype=numpy.int16)
    optimal_shift = imageRegistration.shift_detection(data, 100)
    shift = [-0.5, -0.2]

    numpy.testing.assert_allclose(shift, numpy.round(optimal_shift, 1))


def test_shift_correction00(data):
    """Tests the shift correction of a [0,0] shift."""

    data = imageRegistration.shift_correction(
        data, numpy.outer([0, 0], numpy.arange(3))
    )
    numpy.testing.assert_allclose(data, data, rtol=1e-03)


def test_shift_correction01(data):
    """Tests the shift correction of a [0,1] shift."""

    expected = numpy.array(
        [
            [
                [1, 2, 3, 4, 5],
                [2, 2, 3, 4, 5],
                [3, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
            ],
            [
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 3],
                [1, 2, 3, 4, 5],
            ],
            [
                [8, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
            ],
        ]
    )

    new_data = imageRegistration.shift_correction(
        data, numpy.outer([1, 0], numpy.arange(3))
    )
    numpy.testing.assert_allclose(new_data, expected, rtol=1e-03)


def test_shift_correction10(data):
    """Tests the shift correction of a [1,0] shift."""

    expected = numpy.array(
        [
            [
                [1, 2, 3, 4, 5],
                [2, 2, 3, 4, 5],
                [3, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
            ],
            [
                [5, 1, 2, 3, 4],
                [5, 1, 2, 3, 4],
                [3, 1, 2, 3, 4],
                [5, 1, 2, 3, 4],
                [5, 1, 2, 3, 4],
            ],
            [
                [4, 5, 1, 2, 3],
                [4, 5, 1, 2, 3],
                [4, 5, 1, 2, 3],
                [4, 5, 8, 2, 3],
                [4, 5, 1, 2, 3],
            ],
        ]
    )

    new_data = imageRegistration.shift_correction(
        data, numpy.outer([0, 1], numpy.arange(3))
    )
    numpy.testing.assert_allclose(new_data, expected, rtol=1e-05)


def test_shift_correction11(data):
    """Tests the shift correction of a [1,1] shift."""

    expected = numpy.array(
        [
            [
                [1, 2, 3, 4, 5],
                [2, 2, 3, 4, 5],
                [3, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
            ],
            [
                [5, 1, 2, 3, 4],
                [5, 1, 2, 3, 4],
                [5, 1, 2, 3, 4],
                [3, 1, 2, 3, 4],
                [5, 1, 2, 3, 4],
            ],
            [
                [4, 5, 8, 2, 3],
                [4, 5, 1, 2, 3],
                [4, 5, 1, 2, 3],
                [4, 5, 1, 2, 3],
                [4, 5, 1, 2, 3],
            ],
        ]
    )
    new_data = imageRegistration.shift_correction(
        data, numpy.outer([1, 1], numpy.arange(3))
    )
    numpy.testing.assert_allclose(new_data, expected, rtol=1e-05)


def test_shift_correction_float(data):
    """Tests the shift correction of a [0.1, 0.25] shift between images."""

    expected = [
        [
            [1, 2, 3, 4, 5],
            [2, 2, 3, 4, 5],
            [3, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
        ],
        [
            [1.2595824, 1.7349374, 3.0299857, 3.756984, 4.9321423],
            [1.3387496, 1.6893137, 3.0737815, 3.690435, 5.6077204],
            [1.0840675, 1.836086, 2.9328897, 3.904524, 3.4343739],
            [1.220753, 1.7573147, 3.008505, 3.7896245, 4.600788],
            [1.3292272, 1.6948014, 3.0685136, 3.6984396, 5.5264597],
        ],
        [
            [0.03080546, 1.01156497, 3.30827889, 3.27796478, 5.64089073],
            [2.96707199, 1.77546528, 2.90155831, 3.6526126, 5.10329182],
            [0.03080546, 1.01156497, 3.30827889, 3.27796478, 5.64089073],
            [5.90333852, 2.53936558, 2.49483774, 4.02726042, 4.56569291],
            [5.90333852, 2.53936558, 2.49483774, 4.02726042, 4.56569291],
        ],
    ]

    new_data = imageRegistration.shift_correction(
        data, numpy.outer([0.25, 0.1], numpy.arange(3))
    )
    numpy.testing.assert_allclose(new_data, expected, rtol=1e-05)


@pytest.mark.skipif(scipy is None, reason="scipy is missing")
def test_shift_detection0(tmpdir, first_frame, metadata_dict):
    """Tests the shift detection using only an axis (dimension).
    The shift is only applied to the dimension."""
    data = numpy.ndarray((10, *first_frame.shape), first_frame.dtype)
    shift = numpy.array([0.5, 0.2])
    for j in range(2):
        for i in range(0, 5):
            data[j * 5 + i] = numpy.array(
                numpy.fft.ifftn(
                    scipy.ndimage.fourier_shift(numpy.fft.fftn(first_frame), shift * i)
                ).real
            )
    data = numpy.asanyarray(data, dtype=numpy.int16)
    dataset = utils.createDataset(
        data=data, metadata_dict=metadata_dict, _dir=str(tmpdir)
    )

    dataset.find_dimensions()
    dataset.reshape_data()

    # Detects shift using only images where value 1 of dimension 1 is fixed
    optimal_shift = dataset.find_shift(selected_axis=1)

    shift = [-0.5, -0.2]

    numpy.testing.assert_allclose(shift, numpy.round(optimal_shift, 1))


@pytest.mark.skipif(scipy is None, reason="scipy is missing")
def test_shift_detection1(tmpdir, first_frame, metadata_dict):
    """Tests the shift detection using only an axis (dimension).
    The shift is applied to all the dataset."""
    data = [first_frame]
    shift = [0.5, 0.2]
    for i in range(1, 10):
        data += [
            numpy.fft.ifftn(
                scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
            ).real
        ]
    data = numpy.asanyarray(data, dtype=numpy.int16)
    dataset = utils.createDataset(
        data=data, metadata_dict=metadata_dict, _dir=str(tmpdir)
    )

    dataset.find_dimensions()
    dataset.reshape_data()

    # Detects shift using only images where value 1 of dimension 1 is fixed
    optimal_shift = dataset.find_shift(selected_axis=0)

    shift = [-2.5, -1]

    numpy.testing.assert_allclose(shift, numpy.round(optimal_shift, 1))


@pytest.mark.skipif(scipy is None, reason="scipy is missing")
def test_shift_correction0(tmpdir, first_frame, metadata_dict):
    """Tests the shift correction using only an axis (dimension).
    The shift is only applied to the dimension."""
    data = numpy.ndarray((10, *first_frame.shape), first_frame.dtype)

    shift = numpy.array([0.5, 0.2])
    for j in range(2):
        for i in range(0, 5):
            data[j * 5 + i] = numpy.array(
                numpy.fft.ifftn(
                    scipy.ndimage.fourier_shift(numpy.fft.fftn(first_frame), shift * i)
                ).real
            )

    data = numpy.asanyarray(data, dtype=numpy.int16)
    dataset = utils.createDataset(
        data=data, metadata_dict=metadata_dict, _dir=str(tmpdir)
    )

    dataset.find_dimensions()
    dataset.reshape_data()

    dataset.find_and_apply_shift(selected_axis=1, shift_approach="fft")

    for frame in dataset.data.take(0, 0):
        # Check if the difference between the shifted frames and the sample frame is small enough
        numpy.testing.assert_allclose(dataset.as_array3d(0), frame, atol=6)



import numpy

from darfix.core import array_utils


def test_unique():
    arr = [10, 11, 10, 11, 2, 2, 2, 3, 3, 1]
    return_type = type(numpy.unique(arr))
    result = array_utils.unique(arr)
    assert return_type is type(result)
    assert result.tolist() == [10, 11, 2, 3, 1]

    return_type = type(numpy.unique(arr, return_index=True))
    result = array_utils.unique(arr, return_index=True)
    assert return_type is type(result)
    assert len(result) == 2
    assert result[0].tolist() == [10, 11, 2, 3, 1]
    assert result[1].tolist() == [0, 1, 4, 7, 9]

    return_type = type(numpy.unique(arr, return_index=True, return_counts=True))
    result = array_utils.unique(arr, return_index=True, return_counts=True)
    assert return_type is type(result)
    assert len(result) == 3
    assert result[0].tolist() == [10, 11, 2, 3, 1]
    assert result[1].tolist() == [0, 1, 4, 7, 9]
    assert result[2].tolist() == [2, 2, 3, 2, 1]

    return_type = type(numpy.unique(arr, return_index=True, return_counts=True))
    result = array_utils.unique(arr, return_counts=True)
    assert return_type is type(result)
    assert len(result) == 2
    assert result[0].tolist() == [10, 11, 2, 3, 1]
    assert result[1].tolist() == [2, 2, 3, 2, 1]



__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "26/07/2019"


from silx.gui import qt


class OperationThread(qt.QThread):
    """
    Given a function and a set of arguments, it calls it whenever the thread
    is started.
    """

    def __init__(self, parent, function):
        qt.QThread.__init__(self, parent=parent)
        self.func = function
        self.args = []
        self.kwargs = {}
        self.data = None

    def setArgs(self, *args, **kwargs):
        """
        Function to set the arguments of the function

        :param List args:
        """
        self.args = args
        self.kwargs = kwargs

    def run(self):
        self.data = self.func(*self.args, **self.kwargs)



from __future__ import annotations

from typing import Optional

import numpy
from silx.gui import qt
from silx.gui.plot import Plot1D

from darfix import dtypes


class DataPartitionWidget(qt.QMainWindow):
    sigComputeHistogram = qt.Signal()
    "emit when the user ask for computing the histogram"
    sigPartitionData = qt.Signal()
    "emit when the user ask for filtering the data"
    sigAbort = qt.Signal()
    "emit when user request to abort the processing"
    sigNbBinsChanged = qt.Signal(int)
    "emit when the number of bins has changed"
    sigBottomBinChanged = qt.Signal(int)
    "emit when the bottom bin has changed"
    sigTopBinChanged = qt.Signal(int)
    "emit when the top bin has changed"

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)
        self._dataset = None

        self._plot = Plot1D()

        binsLabel = qt.QLabel("Number of histogram bins:")
        filterRangeLabel = qt.QLabel("Filter range:")
        self._bins = qt.QLineEdit("")
        self._bins.setToolTip(
            "Defines the number of equal-width bins in the given range for the histogram"
        )
        self._bins.setValidator(qt.QIntValidator())
        self._bottomBinsNumber = qt.QLineEdit("")
        self._bottomBinsNumber.setPlaceholderText("First bin")
        self._bottomBinsNumber.setToolTip(
            "Minimum bin to use. It is 0 if nothing is entered."
        )
        self._topBinsNumber = qt.QLineEdit("")
        self._topBinsNumber.setPlaceholderText("Last bin")
        self._topBinsNumber.setToolTip(
            "Maximum bin to use. It is the number of bins if nothing is entered"
        )
        self._bottomBinsNumber.setValidator(qt.QIntValidator())
        self._topBinsNumber.setValidator(qt.QIntValidator())
        self._computeHistogramPB = qt.QPushButton("Compute histogram")
        self._computePartitionPB = qt.QPushButton("Filter")
        self._abortPB = qt.QPushButton("Abort")
        widget = qt.QWidget(parent=self)
        layout = qt.QGridLayout()
        layout.addWidget(binsLabel, 0, 0, 1, 1)
        layout.addWidget(self._bins, 0, 1, 1, 1)
        layout.addWidget(filterRangeLabel, 1, 0, 1, 1)
        layout.addWidget(self._bottomBinsNumber, 1, 1)
        layout.addWidget(self._topBinsNumber, 1, 2)
        layout.addWidget(self._computeHistogramPB, 0, 2, 1, 1)
        layout.addWidget(self._computePartitionPB, 2, 2, 1, 1)
        layout.addWidget(self._abortPB, 3, 2, 1, 1)
        layout.addWidget(self._plot, 3, 0, 1, 3)
        widget.setLayout(layout)
        widget.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self.setCentralWidget(widget)
        self._plot.hide()

        self.setProcessingButtonsEnabled(enable=False)

        # set up
        self._abortPB.hide()

        # connect signal / slot
        self._abortPB.pressed.connect(self.sigAbort)
        self._computeHistogramPB.released.connect(self.sigComputeHistogram)
        self._computePartitionPB.released.connect(self.sigPartitionData)
        self._bins.editingFinished.connect(self._nbBinsChanged)
        self._bottomBinsNumber.editingFinished.connect(self._nbBottomBinChanged)
        self._topBinsNumber.editingFinished.connect(self._nbTopBinChanged)

    def _nbBinsChanged(self):
        self.sigNbBinsChanged.emit(self.getBins())

    def _nbBottomBinChanged(self):
        self.sigBottomBinChanged.emit(self.getBottomBin())

    def _nbTopBinChanged(self):
        self.sigTopBinChanged.emit(self.getTopBin())

    def setDataset(self, dataset):
        if not isinstance(dataset, dtypes.Dataset):
            raise dtypes.DatasetTypeError(dataset)
        self.setBins(dataset.dataset.nframes)
        self.setProcessingButtonsEnabled(enable=True)

    def _showHistogramCallback(self):
        sender = self.sender()
        frames_intensity = sender.data
        self.showHistogram(frames_intensity=frames_intensity)

    def showHistogram(self, frames_intensity: Optional[numpy.array]):
        """
        Plots the frame intensity histogram.
        """
        self._abortPB.hide()
        self._abortPB.setEnabled(True)
        self._computePartitionPB.setEnabled(True)
        self._computeHistogramPB.setEnabled(True)
        self._plot.clear()
        if frames_intensity is not None:
            values, bins = numpy.histogram(
                frames_intensity, numpy.arange(self.getBins())
            )
            self._plot.addHistogram(values, bins, fill=True)
            self._plot.show()

    # bins getter / setter
    def getBins(self) -> Optional[int]:
        text = self._bins.text().replace(" ", "")
        if text == "":
            return None
        else:
            return int(text)

    def setBins(self, value: int | str | None):
        if value in ("", None):
            self._bins.clear()
        else:
            self._bins.setText(str(value))

    def getBottomBin(self) -> Optional[int]:
        text = self._bottomBinsNumber.text().replace(" ", "")
        if text == "":
            return None
        else:
            return int(text)

    def setBottomBin(self, value: int | str | None):
        if value in ("", None):
            self._bottomBinsNumber.clear()
        else:
            self._bottomBinsNumber.setText(str(value))

    def getTopBin(self) -> Optional[int]:
        text = self._topBinsNumber.text().replace(" ", "")
        if text == "":
            return None
        else:
            return int(text)

    def setTopBin(self, value: int | str | None):
        if value in ("", None):
            self._topBinsNumber.clear()
        else:
            self._topBinsNumber.setText(str(value))

    # util function
    def setProcessingButtonsEnabled(self, enable: bool):
        self._abortPB.setEnabled(not enable)
        self._computePartitionPB.setEnabled(enable)
        self._computeHistogramPB.setEnabled(enable)



from silx.gui import qt


class FitComboBox(qt.QComboBox):
    def __init__(self) -> None:
        super().__init__()
        self.addItems(("trf", "dogbox", "lm"))
        self.setItemData(
            self.findText("trf"),
            "Bounded - Trust Region Reflective algorithm, particularly suitable for large sparse problems with bounds. Generally robust method.",
        )
        self.setItemData(
            self.findText("dogbox"),
            "Bounded - dogleg algorithm with rectangular trust regions, typical use case is small problems with bounds. Not recommended for problems with rank-deficient Jacobian",
        )
        self.setItemData(
            self.findText("lm"),
            "Unbounded Levenberg-Marquardt algorithm as implemented in MINPACK. Doesn’t handle bounds and sparse Jacobians. Usually the most efficient method for small unconstrained problems.",
        )

        self.setCurrentText("lm")



from __future__ import annotations

import logging
import os
from typing import Literal

from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict
from silx.io.dictdump import dicttonx

from darfix import dtypes

from ..core.dataset import ImageDataset
from ..core.grainplot import OrientationDistData
from ..core.grainplot import OrientationDistImage
from ..core.grainplot import compute_mosaicity
from ..core.grainplot import generate_grain_maps_nxdict

_logger = logging.getLogger(__file__)


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: dtypes.Dataset
    """ Input dataset containing a stack of images """
    dimensions: tuple[int, int] = (0, 1)
    """Dimension indices to use for the maps. Default is (0, 1), which means the two first dimensions."""
    save_maps: bool = True
    """Whether to save the maps to file. Default is True."""
    filename: str | MissingData = MISSING_DATA
    """Only used if save_maps is True. Filename to save the maps to. Default is 'maps.h5' in the dataset directory."""
    orientation_img_origin: Literal["dims", "center"] = "dims"
    "Origin for the orientation distribution image. Can be 'dims', 'center' or None. Default is 'dims'."


class GrainPlot(
    Task,
    input_model=Inputs,
    output_names=["dataset"],
):
    """Generates and saves maps of Center of Mass, FWHM, Skewness, Kurtosis, Orientation distribution and Mosaicity."""

    def run(self):

        inputs = Inputs(**self.get_input_values())

        default_filename = os.path.join(inputs.dataset.dataset._dir, "maps.h5")
        filename: str = self.get_input_value("filename", default_filename)

        dataset: ImageDataset = inputs.dataset.dataset
        moments = dataset.apply_moments()

        # mosaicity and orientation can only be computed for 2D+ datasets
        if dataset.dims.ndim > 1:
            dimension1, dimension2 = inputs.dimensions

            mosaicity = compute_mosaicity(
                moments,
                x_dimension=dimension1,
                y_dimension=dimension2,
            )

            orientation_dist_data = OrientationDistData(
                dataset,
                x_dimension=dimension1,
                y_dimension=dimension2,
            )
            assert orientation_dist_data is not None

            orientation_dist_image = OrientationDistImage(
                xlabel=orientation_dist_data.x_label,
                ylabel=orientation_dist_data.y_label,
                scale=orientation_dist_data.rgb_key_plot_scale(),
                origin=orientation_dist_data.origin(inputs.orientation_img_origin),
                data=orientation_dist_data.data,
                as_rgb=orientation_dist_data.rgb_key,
                contours=dict(),
            )
        else:
            mosaicity = None
            orientation_dist_image = None

        # Save data if asked
        if inputs.save_maps:
            nxdict = generate_grain_maps_nxdict(
                dataset, mosaicity, orientation_dist_image
            )
            dicttonx(nxdict, filename)

        self.outputs.dataset = dtypes.Dataset(
            dataset=dataset,
            indices=inputs.dataset.indices,
            bg_indices=inputs.dataset.bg_indices,
            bg_dataset=inputs.dataset.bg_dataset,
        )



from __future__ import annotations

import h5py
import numpy
from silx.io.dictdump import h5todict
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset


class Positioners:
    """
    Class to handle positioners in the hdf5

    For now only load method is available.

    Soon a save method will be implemented as well
    """

    _DATASETS_USED_FOR_TRANSFORMATION = ("mainx", "ffz", "obx", "obpitch")
    """A couple of dataset with hard-coded named used for transformation"""

    def __init__(
        self,
        url: str | DataUrl,
        data: None | dict[str, numpy.ndarray] = None,
    ):
        """
        :param url: url to load data
        :param data: data can be given if already loaded
        """
        self._constants = {}
        self._data = {}
        if not isinstance(url, DataUrl):
            url = DataUrl(url)
        self._url = url
        if data is not None:
            self._data = data
        else:
            self.load()

    def load(self):
        """
        Use url to load data
        """
        h5_positioners_dict = {}
        self._constants = {}
        self._data = {}

        with h5py.File(self._url.file_path()) as f:

            if f.attrs.get("publisher", None) == "bliss":
                h5_positioners_dict = extract_bliss_positioners(
                    f[self._url.data_path()]
                )
            else:
                h5_positioners_dict = h5todict(f, self._url.data_path())

        for key, positioner in h5_positioners_dict.items():
            if isinstance(positioner, numpy.ndarray) and positioner.size > 1:
                self._data[key] = positioner.ravel()
            elif key in Positioners._DATASETS_USED_FOR_TRANSFORMATION:
                # The motor value is constant and we need it for the Transformation
                self._constants[key] = positioner

    @property
    def url(self) -> DataUrl:
        return self._url

    @property
    def data(self) -> dict[str, numpy.ndarray]:
        return self._data

    @property
    def constants(self) -> dict[str, float]:
        """Constant for transformation like obpitch..."""
        return {**self._constants}


def extract_bliss_positioners(
    positioners: h5py.Group,
) -> dict[str, numpy.ndarray | float]:
    # Some positioners are detectors and therefore scalars in `positioners`.
    # To detect those we need to check whether the positioner has a group
    # in the instrument group.
    instrument = positioners.parent
    instrument_names = list(instrument)
    datasets = dict()

    for name in positioners:
        dset = None

        if name in instrument_names:
            group = instrument[name]
            NX_class = group.attrs.get("NX_class", None)
            try:
                if NX_class == "NXpositioner":
                    dset = group["value"]
                elif NX_class == "NXdetector":
                    dset = group["data"]
            except KeyError:
                pass

        if dset is None:
            dset = positioners[name]

        if isinstance(dset, h5py.Dataset):
            if (
                dset.ndim == 0 and name in Positioners._DATASETS_USED_FOR_TRANSFORMATION
            ) or (dset.ndim == 1):
                datasets[name] = h5py_read_dataset(dset)
    return datasets



from __future__ import annotations

import logging
from typing import Literal
from typing import Tuple

import numpy
from scipy.optimize import curve_fit

import darfix

from ..math import bivariate_gaussian
from ..math import gaussian
from ..math import trivariate_gaussian

FitMethod = Literal["trf", "lm", "dogbox"]

_ZERO_SUM_RELATIVE_TOLERANCE = 1e-3
""" Relative tolerance used to check if the sum of values equals 0. Skips fit if it is the case."""

_BOUNDS_TOLERANCE = 1e-3
""" Absolute tolerance for to set bounds of fit parameters. The bounds will be set to (min - tol, max + tol) whenever possible."""

_logger = logging.getLogger(__file__)


def fit_1d_rocking_curve(
    y_values: Tuple[numpy.ndarray, numpy.ndarray | None],
    x_values: list | numpy.ndarray | None = None,
    num_points: int | None = None,
    int_thresh: int | None = None,
    method: FitMethod | None = None,
) -> Tuple[numpy.ndarray, numpy.ndarray | list]:
    """
    Fit rocking curve.

    :param y_values: the first element is the dependent data and the second element are
        the moments to use as starting values for the fit
    :param values: The independent variable where the data is measured, optional
    :param num_points: Number of points to evaluate the data on, optional
    :param nt_thresh: Intensity threshold. If not None, only the rocking curves with
        higher ptp (range of values) are fitted, others are assumed to be noise and not important
        data. This parameter is used to accelerate the fit. Optional.

    :returns: If curve was fitted, the fitted curve, else item[0]
    """
    if method is None:
        method = "trf"
    y, moments = y_values
    y = numpy.asanyarray(y)
    x = numpy.asanyarray(x_values) if x_values is not None else numpy.arange(len(y))
    ptp_y = numpy.ptp(y)
    if int_thresh is not None and ptp_y < int_thresh:
        return y, [0, x[0], 0, min(y)]
    if moments is not None:
        p0 = [ptp_y, moments[0], moments[1], min(y)]
    else:
        _sum = sum(y)
        if _sum > 0:
            mean = sum(x * y) / sum(y)
            sigma = numpy.sqrt(sum(y * (x - mean) ** 2) / sum(y))
        else:
            mean, sigma = numpy.nan, numpy.nan
        p0 = [ptp_y, mean, sigma, min(y)]
    if numpy.isnan(mean) or numpy.isnan(sigma):
        return y, p0
    if numpy.isclose(p0[2], 0):
        return y, p0
    if num_points is None:
        num_points = len(y)
    epsilon = 1e-2
    bounds = numpy.array(
        [
            [min(ptp_y, min(y)) - epsilon, min(x) - epsilon, 0, -numpy.inf],
            [max(max(y), ptp_y) + epsilon, max(x) + epsilon, numpy.inf, numpy.inf],
        ]
    )

    p0 = numpy.array(p0)
    p0[p0 < bounds[0]] = bounds[0][p0 < bounds[0]]
    p0[p0 > bounds[1]] = bounds[1][p0 > bounds[1]]
    try:
        pars, cov = curve_fit(
            f=gaussian, xdata=x, ydata=y, p0=p0, bounds=bounds, method=method
        )
        y_gauss = gaussian(numpy.linspace(x[0], x[-1], num_points), *pars)
        y_gauss[numpy.isnan(y_gauss)] = 0
        y_gauss[y_gauss < 0] = 0
        pars[2] *= darfix.config.FWHM_VAL
        return y_gauss, pars
    except RuntimeError:
        p0[2] *= darfix.config.FWHM_VAL
        return y, p0
    except ValueError:
        p0[2] *= darfix.config.FWHM_VAL
        return y, p0


def _get_2d_bounds(
    method: FitMethod, x_values: numpy.array, min_y: float, max_y: float
):
    if method not in ("trf", "dogbox"):
        return (-numpy.inf, numpy.inf)

    min_x0, min_x1 = numpy.min(x_values, axis=1)
    max_x0, max_x1 = numpy.max(x_values, axis=1)

    return (
        [
            min_x0 - _BOUNDS_TOLERANCE,
            min_x1 - _BOUNDS_TOLERANCE,
            -numpy.inf,
            -numpy.inf,
            min_y - _BOUNDS_TOLERANCE,
            -1,
            -numpy.inf,
        ],
        [
            max_x0 + _BOUNDS_TOLERANCE,
            max_x1 + _BOUNDS_TOLERANCE,
            numpy.inf,
            numpy.inf,
            max_y + _BOUNDS_TOLERANCE,
            1,
            numpy.inf,
        ],
    )


def fit_2d_rocking_curve(
    y_values_and_moments: Tuple[numpy.ndarray, numpy.ndarray | None],
    x_values: list | numpy.ndarray,
    shape: Tuple[int, int],
    int_thresh: int | None = None,
    method: FitMethod | None = None,
) -> Tuple[numpy.ndarray, numpy.ndarray | list]:
    if method is None:
        method = "trf"
    y, moments = y_values_and_moments
    y = numpy.asanyarray(y)
    ptp_y = numpy.ptp(y)
    x_values = numpy.asanyarray(x_values)
    sum_y = sum(y)
    if numpy.isclose(sum_y, 0, rtol=_ZERO_SUM_RELATIVE_TOLERANCE):
        return y, [numpy.nan, numpy.nan, numpy.nan, numpy.nan, ptp_y, 0, 0]
    x0_0 = sum(x_values[0] * y) / sum_y
    x1_0 = sum(x_values[1] * y) / sum_y
    x0_alpha = numpy.sqrt(sum(y * (x_values[0] - x0_0) ** 2) / sum_y)
    x1_alpha = numpy.sqrt(sum(y * (x_values[1] - x1_0) ** 2) / sum_y)
    if (
        (int_thresh is not None and ptp_y < int_thresh)
        or x0_alpha == 0
        or x1_alpha == 0
    ):
        return y, [x0_0, x1_0, x0_alpha, x1_alpha, ptp_y, 0, 0]

    try:
        pars, cov = curve_fit(
            f=bivariate_gaussian,
            xdata=x_values,
            ydata=y,
            p0=[x0_0, x1_0, x0_alpha, x1_alpha, ptp_y, 0, 0],
            bounds=_get_2d_bounds(
                method, x_values, min_y=min(y.min(), ptp_y), max_y=max(y.max(), ptp_y)
            ),
            method=method,
        )
        # TODO: Shape is flipped for some reason
        x_gauss = x_values.reshape((2, shape[1], shape[0]))
        y_gauss = bivariate_gaussian(x_gauss, *pars)
        pars[2] *= darfix.config.FWHM_VAL
        pars[3] *= darfix.config.FWHM_VAL
        return y_gauss.ravel(), pars
    except RuntimeError:
        return y, [
            x0_0,
            x1_0,
            darfix.config.FWHM_VAL * x0_alpha,
            darfix.config.FWHM_VAL * x1_alpha,
            ptp_y,
            0,
            0,
        ]


def _get_3d_bounds(
    method: FitMethod, x_values: numpy.ndarray, min_y: float, max_y: float
):
    """
    Computes bounds for the curve fit of trivariate gaussian.

    The bounds are for the following parameters (in order):
        - x0_0
        - x1_0
        - x2_0
        - sigma_x0
        - sigma_x1
        - sigma_x2
        - c10
        - c12
        - c20
        - amplitude
        - background
    """
    if method == "lm":
        # `lm` cannot handle bounds: no point in calculating them
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html#scipy.optimize.least_squares
        return (-numpy.inf, numpy.inf)

    mins = numpy.min(x_values, axis=1) - _BOUNDS_TOLERANCE
    maxs = numpy.max(x_values, axis=1) + _BOUNDS_TOLERANCE

    return (
        [
            mins[0],
            mins[1],
            mins[2],
            -numpy.inf,
            -numpy.inf,
            -numpy.inf,
            -1,
            -1,
            -1,
            min_y - _BOUNDS_TOLERANCE,
            -numpy.inf,
        ],
        [
            maxs[0],
            maxs[1],
            maxs[2],
            numpy.inf,
            numpy.inf,
            numpy.inf,
            1,
            1,
            1,
            max_y + _BOUNDS_TOLERANCE,
            numpy.inf,
        ],
    )


def fit_3d_rocking_curve(
    y_values_and_moments: Tuple[numpy.ndarray, numpy.ndarray | None],
    x_values: numpy.ndarray,
    shape: Tuple[int, int, int],
    int_thresh: int | None = None,
    method: FitMethod | None = None,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    if method is None:
        method = "trf"

    y, moments = y_values_and_moments
    ptp_y = numpy.ptp(y)
    sum_y = sum(y)

    default_params = [
        numpy.nan,
        numpy.nan,
        numpy.nan,
        numpy.nan,
        numpy.nan,
        numpy.nan,
        0,
        0,
        0,
        ptp_y,
        0,
    ]
    if numpy.isclose(sum_y, 0, rtol=_ZERO_SUM_RELATIVE_TOLERANCE):
        return y, default_params

    X0 = numpy.sum(x_values * y, axis=1) / sum_y
    X_sigma = numpy.sqrt(
        numpy.sum(y * (x_values - X0[:, numpy.newaxis]) ** 2, axis=1) / sum_y
    )

    default_params = [*X0, *(X_sigma * darfix.config.FWHM_VAL), 0, 0, 0, ptp_y, 0]
    if (int_thresh is not None and ptp_y < int_thresh) or numpy.any(X0 == 0):
        return y, default_params

    try:
        fit_params, cov = curve_fit(
            f=trivariate_gaussian,
            xdata=x_values,
            ydata=y,
            p0=[*X0, *X_sigma, 0, 0, 0, ptp_y, 0],
            method=method,
            bounds=_get_3d_bounds(
                method, x_values, min_y=min(y.min(), ptp_y), max_y=max(y.max(), ptp_y)
            ),
        )
    except (RuntimeError, ValueError) as e:
        _logger.warning(
            f"Encountered the following error while fitting rocking curves: '{e}'"
        )
        return y, default_params
    else:
        x_gauss = x_values.reshape((3, shape[2], shape[1], shape[0]))
        y_gauss = trivariate_gaussian(x_gauss, *fit_params)
        fit_params[3] *= darfix.config.FWHM_VAL
        fit_params[4] *= darfix.config.FWHM_VAL
        fit_params[5] *= darfix.config.FWHM_VAL
        return y_gauss.ravel(), fit_params



import numpy

from darfix.dtypes import Dataset
from darfix.tasks.copy import DataCopy
from darfix.tests.utils import createDataset


def test_data_copy():
    dataset = Dataset(
        dataset=createDataset(data=numpy.linspace(1, 10, 100).reshape(10, 5, 2)),
        indices=None,
        bg_dataset=numpy.ones(
            100,
        ).reshape(10, 5, 2),
        bg_indices=None,
    )

    task = DataCopy(
        inputs={
            "dataset": dataset,
        }
    )
    task.run()

    assert id(dataset) != id(task.outputs.dataset)
    assert id(dataset.dataset) != id(task.outputs.dataset.dataset)
    assert id(dataset.bg_dataset) != id(task.outputs.dataset.bg_dataset)
    assert dataset.dataset.data.shape == dataset.bg_dataset.shape
    numpy.testing.assert_array_equal(
        dataset.dataset.data, task.outputs.dataset.dataset.data
    )
    numpy.testing.assert_array_equal(
        dataset.bg_dataset, task.outputs.dataset.bg_dataset
    )



import os

from ewoksorange.tests.conftest import qtapp  # noqa F401

from darfix.gui.utils.qsignalspy import QSignalSpy
from orangecontrib.darfix.widgets.concatenateHDF5 import ConcatenateWidgetOW

from ...tasks.test_hdf5_scans_concatenation import create_scans


def test_ConcatenateWidgetOW(tmp_path, qtapp):  # noqa F811
    """
    test the ConcatenateWidgetOW

    Define inputs from the GUI and launch concatenation.
    Make sure the expected output file exists (but doesn't check the content. The Ewoks task should check this).
    """
    window = ConcatenateWidgetOW()

    assert window.get_task_input_values() == {
        "detector_data_path": r"{scan}/measurement/{detector}",
        "positioners_group_path": r"{scan}/instrument/positioners",
        "overwrite": True,
    }

    test_dir = tmp_path / "test_ConcatenateWidgetOW"
    test_dir.mkdir()
    raw_data_file = os.path.join(str(test_dir), "scan_to_concateante.h5")
    create_scans(file_path=raw_data_file, n_scan=3)

    concat_widget = window._widget._mainWidget
    concat_widget.setOverwrite(False)
    concat_widget.setInputFile(raw_data_file)
    output_file = os.path.join(str(test_dir), "scan_to_concateante_darfix_concat.h5")
    assert window.get_task_input_values() == {
        "detector_data_path": r"{scan}/measurement/{detector}",
        "positioners_group_path": "{scan}/instrument/positioners",
        "overwrite": True,
        "input_file": raw_data_file,
        "output_file": output_file,
    }
    assert not os.path.exists(output_file)

    waiter = QSignalSpy(window.task_executor.finished)
    window.execute_ewoks_task()
    waiter.wait(5000)
    assert os.path.exists(output_file)
    assert window.task_executor.succeeded

    concat_widget.setDetectorPath("/first/path")
    concat_widget.setPositionersGroupPath("/second/path")
    assert window.get_task_input_values() == {
        "overwrite": True,
        "input_file": raw_data_file,
        "output_file": output_file,
        "detector_data_path": "/first/path",
        "positioners_group_path": "/second/path",
    }
    concat_widget.setSelectedEntries(("/1.1", "3.1"))
    assert window.get_task_input_values() == {
        "overwrite": True,
        "entries_to_concatenate": ("/1.1", "/3.1"),
        "input_file": raw_data_file,
        "output_file": output_file,
        "detector_data_path": "/first/path",
        "positioners_group_path": "/second/path",
    }



from __future__ import annotations

from typing import Sequence

from ewokscore import Task
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict

from ..dtypes import Dataset
from ..dtypes import DatasetTypeError


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: Dataset
    """ Input dataset containing a stack of images """
    dimension: Sequence[int]
    """Dimensions indices to project the data onto."""


class Projection(Task, input_model=Inputs, output_names=["dataset"]):
    """
    Removes one dimension by projecting (summing) all images in this dimension.

    Details in https://gitlab.esrf.fr/XRD/darfix/-/issues/37
    """

    def run(self):
        dataset = self.inputs.dataset

        if not isinstance(dataset, Dataset):
            raise DatasetTypeError(dataset)

        darfix_dataset = dataset.dataset
        indices = dataset.indices
        dimension: Sequence[int] = self.inputs.dimension

        darfix_dataset = darfix_dataset.project_data(
            dimension=dimension, indices=indices
        )

        self.outputs.dataset = Dataset(
            dataset=darfix_dataset,
            indices=indices,
            bg_indices=dataset.bg_indices,
            bg_dataset=dataset.bg_dataset,
        )



from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread

from darfix.tasks.copy import DataCopy


class DataCopyWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=DataCopy):
    """
    Widget that creates a new dataset from a given one, and copies its data.
    """

    name = "data copy"
    icon = "icons/copy.svg"
    want_main_area = False



import numpy
import pytest

from darfix.processing import imageOperations


@pytest.fixture
def data():
    return numpy.array(
        [
            [1, 1, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 6, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
        ],
        dtype=numpy.uint16,
    )


@pytest.fixture
def dark():
    return numpy.array(
        [
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
        ],
        dtype=numpy.uint16,
    )


def test_background_subtraction(data, dark):
    """Tests background subtraction function"""
    expected = numpy.array(
        [
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 3, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
        ],
        dtype=numpy.uint16,
    )

    imageOperations.background_subtraction(data, dark)
    numpy.testing.assert_array_equal(expected, data)


def test_hot_pixel_removal(data):
    """Tests the hot pixel removal in stack of arrays"""
    expected = numpy.array(
        [
            [1, 1, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
        ],
        dtype=numpy.uint16,
    )

    imageOperations.hot_pixel_removal(data)
    numpy.testing.assert_array_equal(expected, data)


def test_threshold_removal(data):
    """Tests the threshold of the data"""

    expected = numpy.array(
        [
            [0, 0, 3, 4, 0],
            [0, 2, 3, 4, 0],
            [0, 2, 0, 4, 0],
            [0, 2, 3, 4, 0],
            [0, 2, 3, 4, 0],
        ],
        dtype=numpy.uint16,
    )

    imageOperations.threshold_removal(data, 2, 4)

    numpy.testing.assert_array_equal(expected, data)



from __future__ import annotations

from ewoksorange.gui.parameterform import block_signals
from silx.gui import qt

from darfix.core.dimension import AcquisitionDims


class FilterByDimensionWidget(qt.QWidget):
    """
    Widget to choose a dimension from a dict and choose the value to filter
    the data. It can be included in other widget like StackView to filter the
    stack.
    """

    axisChanged = qt.Signal(int)
    filterChanged = qt.Signal(tuple)
    stateDisabled = qt.Signal()

    def __init__(self, parent=None, vertical=True):

        qt.QWidget.__init__(self, parent)

        if vertical:
            self.setLayout(qt.QVBoxLayout())
        else:
            self.setLayout(qt.QHBoxLayout())

        self.dimensions: AcquisitionDims | None = None

        self._axisComboBox = qt.QComboBox()
        self._axisComboBox.setDisabled(True)
        self._filter_by_dimension_checkbox = qt.QCheckBox("Filter by dimension", None)
        self._filter_by_dimension_checkbox.setDisabled(True)

        self.layout().addWidget(self._axisComboBox)
        self.layout().addWidget(self._filter_by_dimension_checkbox)

        self._axisComboBox.currentIndexChanged.connect(self.axisChanged)
        self._filter_by_dimension_checkbox.toggled.connect(self._updateState)

    def isFilteredByDim(self) -> bool:
        return self._filter_by_dimension_checkbox.isChecked()

    def setFilterByDim(self, filter: bool):
        return self._filter_by_dimension_checkbox.setChecked(filter)

    def getCurrentAxisName(self) -> str:
        """Axis name"""
        return self._axisComboBox.currentText()

    def getCurrentDatasetAxis(self) -> int:
        """Axis according to darfix dataset convention"""
        return self._axisComboBox.currentIndex()

    def getCurrentDimensionAxis(self) -> int:
        """Axis according to darfix dimension convention"""
        return self.dimensions.ndim - 1 - self._axisComboBox.currentIndex()

    def setDimensions(self, dimensions: AcquisitionDims) -> None:
        """Initialize widget with AcquisitionDims instance"""
        with block_signals(
            self._axisComboBox
        ):  # Do not fire valueChanged or filterChanged when setting up the widget
            self.dimensions = dimensions
            self._axisComboBox.clear()
            self._axisComboBox.addItems(self.dimensions.get_names()[::-1])
            self._axisComboBox.setDisabled(True)
            self._filter_by_dimension_checkbox.setChecked(False)
            self._filter_by_dimension_checkbox.setEnabled(self.dimensions.ndim > 1)

    def _updateState(self, checked: bool) -> None:
        """
        Updates the state of the widget.

        :param checked: If True, the widget emit signal with the current value of combo box / sliders. Else,
                    a disabled signal is emitted.

        """
        self._axisComboBox.setEnabled(checked)

        if checked:
            self.axisChanged.emit(self._axisComboBox.currentIndex())
        else:
            self.stateDisabled.emit()



from __future__ import annotations

from silx.gui import qt

from ..pixel_sizes import PixelSize


class RSMWidget(qt.QWidget):
    """
    Widget to transform axes of RSM data
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._rotate = False
        self._moments = None
        self._pixelSize = None

        layout = qt.QGridLayout()

        pixelSizeLabel = qt.QLabel("Pixel size:")
        self._pixelSizeCB = qt.QComboBox()
        self._pixelSizeCB.addItems(PixelSize.names())
        self._rotateCB = qt.QCheckBox("Rotate RSM", self)
        layout.addWidget(pixelSizeLabel, 0, 0)
        layout.addWidget(self._pixelSizeCB, 0, 1)
        layout.addWidget(self._rotateCB, 1, 1)
        self.setLayout(layout)

    @property
    def pixelSize(self):
        return self._pixelSizeCB.currentText()

    @pixelSize.setter
    def pixelSize(self, pixelSize):
        self._pixelSizeCB.setCurrentText(str(pixelSize))

    @property
    def rotate(self):
        return self._rotateCB.isChecked()

    @rotate.setter
    def rotate(self, rotate):
        self._rotateCB.setChecked(rotate)



import numpy
import pytest

from darfix.core import rocking_curves
from darfix.core.utils import NoDimensionsError
from darfix.processing.rocking_curves import fit_1d_rocking_curve

from .utils import createHDF5Dataset2D
from .utils import createHDF5Dataset3D


def test_generator():
    """Tests the correct creation of a generator without moments"""
    data = numpy.random.random(size=(3, 10, 10))
    g = rocking_curves._rocking_curves_per_px(data)

    img, moment = next(g)
    assert moment is None
    numpy.testing.assert_array_equal(img, data[:, 0, 0])


def test_generator_with_moments():
    """Tests the correct creation of a generator with moments"""
    data = numpy.random.random(size=(3, 10, 10))
    moments = numpy.ones((3, 10, 10))
    g = rocking_curves._rocking_curves_per_px(data, moments)

    img, moment = next(g)
    numpy.testing.assert_array_equal(moment, moments[:, 0, 0])
    numpy.testing.assert_array_equal(img, data[:, 0, 0])


def test_fit_1d_rocking_curve():
    """Tests the correct fit of a rocking curve"""

    samples = numpy.random.normal(size=10000) + numpy.random.random(10000)

    y, bins = numpy.histogram(samples, bins=100)

    y_pred, pars = fit_1d_rocking_curve((y, None))
    rss = numpy.sum((y - y_pred) ** 2)
    tss = numpy.sum((y - y.mean()) ** 2)
    r2 = 1 - rss / tss

    assert r2 > 0.9
    assert len(pars) == 4


def test_fit_1d_data():
    """Tests the new data has same shape as initial data"""
    data = numpy.random.random(size=(3, 10, 10))
    new_data, maps = rocking_curves.fit_1d_data(data)

    assert new_data.shape == data.shape
    assert len(maps) == 4
    assert maps[0].shape == data[0].shape


def test_apply_fit_on_2d_dataset():
    data = numpy.random.randint(2, 1000, size=1500).reshape((3, 5, 10, 10))
    dataset = createHDF5Dataset2D(data)
    with pytest.raises(NoDimensionsError):
        fit_dataset, maps = dataset.apply_fit()
    dataset.find_dimensions()
    indices = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    fit_dataset, maps = dataset.apply_fit(indices=indices)

    assert dataset.dims.ndim == 2
    assert len(maps) == 7
    assert fit_dataset.nframes == dataset.nframes
    numpy.testing.assert_equal(fit_dataset.as_array3d(10), dataset.as_array3d(10))


def test_apply_fit_on_3d_dataset():
    data = numpy.random.randint(2, 1000, size=6000).reshape((3, 4, 5, 10, 10))
    dataset = createHDF5Dataset3D(data)
    with pytest.raises(NoDimensionsError):
        fit_dataset, maps = dataset.apply_fit()
    dataset.find_dimensions()
    assert dataset.dims.ndim == 3
    fit_dataset, maps = dataset.apply_fit()

    assert len(maps) == 11
    assert fit_dataset.nframes == dataset.nframes



import numpy
import pytest

from darfix.core.imageRegistration import apply_opencv_shift
from darfix.tests.utils import createHDF5Dataset2D


@pytest.fixture
def two_motors_dataset():
    # 2d daset like in shift correction example, A square shifted by (3,2) along the axis 1
    data = numpy.zeros(150000).reshape((10, 5, 50, 60))
    for i in range(10):
        data[i, :, 3 * i + 10 : 3 * i + 20, 2 * i + 10 : 2 * i + 20] = 1
    return createHDF5Dataset2D(data)


def test_find_shift(two_motors_dataset):
    """Tests the shift detection with dimensions and indices"""
    dataset = two_motors_dataset

    dataset.find_dimensions()
    dataset.reshape_data()

    # Axis 0 : dim size is 5
    assert dataset.dims[0].size == 5
    # Axis 1 : dim size is 10
    assert dataset.dims[1].size == 10

    shift = dataset.find_shift(selected_axis=0)

    # dimension shift is 3, 2 for axis 0
    numpy.testing.assert_allclose(shift, (-3, -2))

    # dimension shift is 0, 0 for axis 1

    shift = dataset.find_shift(selected_axis=1)

    numpy.testing.assert_allclose(shift, (0, 0))


def test_apply_shift(two_motors_dataset):
    """Tests the shift correction with dimensions and indices"""
    dataset = two_motors_dataset

    dataset.find_dimensions()
    dataset.reshape_data()

    # Axis 0 : dim size is 5
    assert dataset.dims[0].size == 5
    # Axis 1 : dim size is 10
    assert dataset.dims[1].size == 10

    input_img = dataset.as_array3d(4).copy()
    # Check fit is well applied at index 4
    result_idx4 = apply_opencv_shift(input_img, shift=(0.5 * 4, 0.5 * 4))

    dataset.apply_shift(shift=(0.5, 0.5), axis=1, shift_approach="fft")

    numpy.testing.assert_raises(
        AssertionError,
        numpy.testing.assert_array_equal,
        input_img,
        dataset.as_array3d(4),
    )
    numpy.testing.assert_allclose(result_idx4, dataset.as_array3d(4))


def test_find_shift_along_dimension(two_motors_dataset):
    """Tests the shift detection along a dimension"""
    dataset = two_motors_dataset

    dataset.find_dimensions()
    dataset.reshape_data()

    shift = dataset.find_shift(selected_axis=1)
    assert len(shift) == 2
    shift = dataset.find_shift(selected_axis=0)
    assert len(shift) == 2


def test_apply_shift_along_dimension(two_motors_dataset):
    """Tests the shift correction with dimensions and indices"""
    dataset = two_motors_dataset

    dataset.find_dimensions()
    dataset.reshape_data()

    # Axis 0 : dim size is 5
    assert dataset.dims[0].size == 5
    # Axis 1 : dim size is 10
    assert dataset.dims[1].size == 10

    first_img = dataset.as_array3d()[0].copy()

    last_img = dataset.as_array3d()[-1].copy()

    dataset.apply_shift(shift=numpy.array([-3, -2]), axis=0)

    # Test if shift correction is successfull
    numpy.testing.assert_array_almost_equal(first_img, dataset.as_array3d()[0])
    numpy.testing.assert_array_almost_equal(first_img, dataset.as_array3d()[24])
    numpy.testing.assert_array_almost_equal(first_img, dataset.as_array3d()[37])
    numpy.testing.assert_array_almost_equal(first_img, dataset.as_array3d()[49])
    numpy.testing.assert_raises(
        AssertionError,
        numpy.testing.assert_array_equal,
        last_img,
        dataset.as_array3d()[-1],
    )



"""Actions associated to different configuration levels (required, advanced and optional)"""

from silx.gui import icons
from silx.gui import qt


class RequiredConfigurationAction(qt.QAction):
    """
    Action to display the 'required' options only
    """

    def __init__(self, parent):
        super().__init__("required options", parent)
        self.__icon = icons.getQIcon("darfix:gui/icons/required_settings")
        self.setIcon(self.__icon)
        self.setIconVisibleInMenu(True)
        self.setCheckable(True)
        self.setToolTip(self.tooltip())

    def tooltip(self):
        return "Show only required options"

    def icon(self):
        return self.__icon

    def text(self):
        return "Required configuration"


class OptionalConfigurationAction(qt.QAction):
    """
    Action to display the 'optional' options only
    """

    def __init__(self, parent):
        super().__init__("optional options", parent)
        self.__icon = icons.getQIcon("darfix:gui/icons/optional_settings")
        self.setIcon(self.__icon)
        self.setIconVisibleInMenu(True)
        self.setCheckable(True)
        self.setToolTip(self.tooltip())

    def tooltip(self):
        return "Show required and optional options. Not any advanced options displayed"

    def icon(self):
        return self.__icon

    def text(self):
        return "Optional configuration"


class AdvancedConfigurationAction(qt.QAction):
    """
    Action to display the 'advanced' / expert options
    """

    def __init__(self, parent):
        super().__init__("advanced options", parent)
        self.__icon = icons.getQIcon("darfix:gui/icons/advanced_settings")
        self.setIcon(self.__icon)
        self.setIconVisibleInMenu(True)
        self.setCheckable(True)
        self.setToolTip(self.tooltip())

    def tooltip(self):
        return "Show all existing options"

    def icon(self):
        return self.__icon

    def text(self):
        return "Advanced configuration"



import os

import h5py
import numpy
import pytest
from ewoksorange.tests.conftest import qtapp  # noqa F401

from darfix.gui.concatenate_scans import ConcatenateHDF5Scans
from darfix.gui.concatenate_scans import ConcatenateScansWidget
from darfix.gui.concatenate_scans import HDF5ConcatenateWindow
from darfix.gui.concatenate_scans import HDF5EntriesWidget
from darfix.gui.configuration.level import ConfigurationLevel


@pytest.mark.parametrize(
    "config_level", (ConfigurationLevel.ADVANCED, ConfigurationLevel.REQUIRED)
)
def test_HDF5ConcatenateWindow_config_level(config_level, qtapp):  # noqa F811
    """test the HDF5ConcatenateWindow different configuration level"""
    window = HDF5ConcatenateWindow()
    widget = window.getConcatenateScansWidget()
    widget.setConfigurationLevel(config_level)
    required_widgets = (
        widget._inputFileSelector,
        widget._scanEntries,
    )
    advanced_widgets = (
        widget._outputFileSelector,
        widget._autoUpdate,
        widget._positionerDataPath,
        widget._detectorDataPath,
        widget._overwriteOutput,
    )
    window.show()
    for widget in required_widgets:
        assert widget.isVisible()

    for widget in advanced_widgets:
        if config_level >= ConfigurationLevel.ADVANCED:
            assert widget.isVisible()
        else:
            assert not widget.isVisible()


def test_ConcatenateScansWidget(tmp_path, qtapp):  # noqa F811
    """
    test of the ConcatenateScansWidget widget.

    Make sure it is correctly updated when a file is provided
    """
    test_dir = tmp_path / "test_concatenate_widget"
    test_dir.mkdir()

    test_file = os.path.join(test_dir, "test.hdf5")
    n_scan = 4
    with h5py.File(test_file, mode="w") as h5f:
        for i_scan in range(n_scan):
            # measurement
            h5f[f"{i_scan + 1}.1/measurement/detector"] = numpy.linspace(
                0, 100, 1000
            ).reshape((10, 10, 10))
            for pos_name, pos_value in {
                "alpha": 1,
                "beta": numpy.arange(0 + i_scan, 10 + i_scan, 1),
            }.items():
                h5f[f"{i_scan + 1}.1/instrument/positioners/{pos_name}"] = pos_value
        # add a dataset to make sure those are filtered
        h5f["dataset"] = "test"

    widget = ConcatenateScansWidget()
    widget.setInputFile(test_file)
    assert widget.getSelectedEntries() == tuple([f"/{i+1}.1" for i in range(n_scan)])
    assert widget.getOutputFile() == os.path.join(test_dir, "test_darfix_concat.hdf5")
    assert widget.getInputFile() == test_file
    assert (
        widget.getDetectorDataPathSelection().getPattern()
        == ConcatenateHDF5Scans.DEFAULT_DETECTOR_DATA_PATH
    )
    assert (
        widget.getDetectorDataPathSelection().getExample()
        == "/1.1/measurement/detector"
    )
    assert (
        widget.getMetadataPathSelection().getPattern()
        == ConcatenateHDF5Scans.DEFAULT_POSITIONERS_DATA_PATH
    )
    assert (
        widget.getMetadataPathSelection().getExample() == "/1.1/instrument/positioners"
    )


def test_HDF5EntriesWidget(qtapp):  # noqa F811
    """
    test of the HDF5EntriesWidget widget

    common test on selection / unselection of items
    """
    widget = HDF5EntriesWidget()
    n_scan = 5
    widget.setEntries([f"{i+1}.1" for i in range(n_scan)])
    assert len(widget.getEntries()) == n_scan
    assert len(widget.getSelectedEntries()) == n_scan

    widget.setSelectedEntries(("1.1", "2.1"))
    assert widget.getSelectedEntries() == ("1.1", "2.1")

    widget.selectAll()
    assert widget.getSelectedEntries() == tuple([f"{i+1}.1" for i in range(n_scan)])

    widget.unselectAll()
    assert widget.getSelectedEntries() == ()



from __future__ import annotations

import logging
from enum import Enum

import numpy
from ewokscore.missing_data import is_missing_data
from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread

from darfix import dtypes
from darfix.gui.roiSelectionWidget import ROISelectionWidget
from darfix.tasks.roi import RoiSelection

_logger = logging.getLogger(__name__)


class _ROIBehavior(Enum):
    """Define the different behavior we can have regarding ROI"""

    FIT_TO_DATASET = "fit-to-dataset"
    """When a ROI is applied we want to update the ROI to the shape to show the user that the ROI has been applied"""
    PROPOSE_USER_ROI = "propose-ROI-to-user"
    """When a new dataset is set (or when resetting) we want to propose to the user a subset of the frame (one fifth of the current frame shape)"""


class RoiSelectionWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=RoiSelection):
    name = "roi selection"
    icon = "icons/roi.png"
    want_main_area = True
    want_control_area = False

    _ewoks_inputs_to_hide_from_orange = ("roi_origin", "roi_size")

    def __init__(self):
        super().__init__()

        self._widget = ROISelectionWidget(parent=self)
        self.mainArea.layout().addWidget(self._widget)
        self._widget.sigApply.connect(self._updateActiveDataset)
        self._widget.sigValidate.connect(self._validateResult)
        self._widget.sigReset.connect(self._reset)

        self._update_dataset: dtypes.Dataset = None
        """Dataset updated when the users press 'apply'"""
        self._dataset: dtypes.Dataset = None
        """Original input dataset"""

        self._nextRoiInteractionBehavior: _ROIBehavior = _ROIBehavior.PROPOSE_USER_ROI
        """When a new output is received we want either to:
        * propose a ROI fitting the dataset shape (if a ROI - subset of the dataset - has been applied)
        * propose a ROI on a subset of the dataset (in the case a new dataset has been received)
        """

    def _validateResult(self):
        self._nextRoiInteractionBehavior = _ROIBehavior.PROPOSE_USER_ROI
        self.propagate_downstream()
        self.accept()

    def handleNewSignals(self):
        dataset: dtypes.Dataset = self.get_task_input_value("dataset")

        if is_missing_data(dataset):
            return

        self._dataset = self._update_dataset = dataset

        self._widget.setDataset(dataset)
        self.open()
        self._nextRoiInteractionBehavior = _ROIBehavior.PROPOSE_USER_ROI

        if not self._tryRecoverLastRoi():
            self._resetTaskAndROI(dataset=dataset)

    def _updateActiveDataset(self):

        self._widget.clampRoiToDataset(self._update_dataset.dataset)
        roi = self._widget.getRoi()

        self.set_default_input("roi_origin", roi.getOrigin().tolist())
        self.set_default_input("roi_size", roi.getSize().tolist())
        self._nextRoiInteractionBehavior = _ROIBehavior.FIT_TO_DATASET
        self.execute_ewoks_task_without_propagation()
        self._widget.enableApplyingROI(False)

    def _updateROI(self):
        if self._nextRoiInteractionBehavior is _ROIBehavior.FIT_TO_DATASET:
            # Apply
            self._widget.enableValidation(True)
            self._widget._fitROIToDataset(self._update_dataset.dataset)
        elif self._nextRoiInteractionBehavior is _ROIBehavior.PROPOSE_USER_ROI:
            # Reset
            self._widget.enableValidation(False)
            self._widget._setROIForNewDataset(self._update_dataset.dataset)
        else:
            raise NotImplementedError(
                f"Behavior not defined ({self._nextRoiInteractionBehavior})"
            )

    def task_output_changed(self):
        super().task_output_changed()
        self._update_dataset = self.get_task_output_value("dataset")

        if is_missing_data(self._update_dataset):
            _logger.error("Roi application failed (or cancelled)")
            return

        self._updateROI()

        self._widget.setStack(self._update_dataset.dataset)
        # update the input dataset for next iteration
        self.set_dynamic_input("dataset", self._update_dataset)

    def _reset(self):
        self._update_dataset = self._dataset

        self._nextRoiInteractionBehavior = _ROIBehavior.PROPOSE_USER_ROI
        self._widget.setDataset(self._update_dataset)
        self._resetTaskAndROI(dataset=self._update_dataset)

    def _tryRecoverLastRoi(self) -> bool:
        """Look at saved default input in .ows and try propose the roi on the current dataset."""
        origin = self.get_task_input_value("roi_origin")
        size = self.get_task_input_value("roi_size")

        if (
            is_missing_data(origin)
            or is_missing_data(size)
            or len(origin) != 2
            or len(size) != 2
        ):
            return False

        origin = numpy.array(origin)
        size = numpy.array(size)
        self._widget.setRoi(size=size, origin=origin)
        self._widget.enableValidation(False)
        self._widget.enableApplyingROI(True)
        return True

    def _resetTaskAndROI(self, dataset: dtypes.Dataset):
        """
        Reset the original dataset (last dataset received) and the ROI.

        When receiving a new dataset or resetting it, we need to process the task to
        retrieve the task output dataset when the user presses ok.
        To speed up process in this specific case (ROI match frame shape), the task will simply propagate the original dataset.
        """
        self.set_default_input("roi_origin", (0, 0))
        self.set_default_input("roi_size", dataset.dataset.frame_shape)
        self.set_dynamic_input("dataset", dataset)
        self.execute_ewoks_task_without_propagation()

        self._widget.enableValidation(False)
        self._widget.enableApplyingROI(True)



from __future__ import annotations

from enum import Enum

import numpy
from ewoksorange.gui.parameterform import block_signals
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D
from silx.gui.plot.items import ImageBase
from silx.image.marchingsquares import find_contours

from ...core.grainplot import MultiDimMomentType
from ...core.grainplot import OrientationDistData
from ...core.grainplot import compute_mosaicity
from ...dtypes import ImageDataset
from ..chooseDimensions import ChooseDimensionWidget
from ..utils.axis_type_combobox import AxisTypeComboBox
from .utils import MapType
from .utils import add_image_with_transformation


class OriDistButtonIds(Enum):
    DATA = 0
    COLOR_KEY = 1


class MosaicityWidget(qt.QWidget):
    """Widget to display and explore mosaicity plots"""

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)

        self._dataset: ImageDataset | None = None
        self._orientation_dist_data: OrientationDistData | None = None
        self._mosaicity: numpy.ndarray | None = None
        self.contours = {}

        layout = qt.QVBoxLayout()
        mosaicity_layout = qt.QHBoxLayout()

        self._chooseDimensionWidget = ChooseDimensionWidget(
            self, vertical=False, values=False
        )
        self._chooseDimensionWidget.valueChanged.connect(
            self._computeMosaicityAndOriDist
        )
        layout.addWidget(self._chooseDimensionWidget)

        self._mosaicityPlot = Plot2D(parent=self)
        self._mosaicityPlot.getColorBarWidget().hide()
        mosaicity_layout.addWidget(self._mosaicityPlot)
        self._levelsWidget = qt.QWidget()

        self._contoursPlot = Plot2D(parent=self)
        self._contoursPlot.getColorBarWidget().hide()

        self._levelsLE = qt.QLineEdit("20")
        self._levelsLE.setToolTip("Number of levels to use when finding the contours")
        self._levelsLE.setValidator(qt.QIntValidator())

        self._computeContoursB = qt.QPushButton("Find contours")
        self._computeContoursB.clicked.connect(self._computeContours)

        self._axisTypeCB = AxisTypeComboBox()
        self._axisTypeCB.currentIndexChanged.connect(self._plotOrientationData)

        self._oriDistTypeButtons = qt.QButtonGroup()
        oriDistButtonsLayout = qt.QHBoxLayout()
        dataButton = qt.QRadioButton(text="Data")
        self._oriDistTypeButtons.addButton(dataButton, id=OriDistButtonIds.DATA.value)
        oriDistButtonsLayout.addWidget(dataButton)
        colorKeyButton = qt.QRadioButton(text="Color key")
        colorKeyButton.setChecked(True)
        oriDistButtonsLayout.addWidget(colorKeyButton)
        self._oriDistTypeButtons.addButton(
            colorKeyButton, id=OriDistButtonIds.COLOR_KEY.value
        )
        self._oriDistTypeButtons.buttonClicked.connect(self._plotOrientationData)

        levelsLayout = qt.QGridLayout()
        levelsLayout.addWidget(qt.QLabel("Number of levels:"), 0, 0, 1, 1)
        levelsLayout.addWidget(self._levelsLE, 0, 1, 1, 1)
        levelsLayout.addWidget(qt.QLabel("Axis values"), 0, 2, 1, 1)
        levelsLayout.addWidget(self._axisTypeCB, 0, 3, 1, 1)
        levelsLayout.addWidget(self._computeContoursB, 1, 0, 1, 2)
        levelsLayout.addLayout(oriDistButtonsLayout, 2, 0, 1, 4)
        levelsLayout.addWidget(self._contoursPlot, 3, 0, 1, 4)
        self._levelsWidget.setLayout(levelsLayout)
        mosaicity_layout.addWidget(self._levelsWidget)

        layout.addLayout(mosaicity_layout)
        self.setLayout(layout)

    @property
    def dimension1(self) -> int:
        try:
            return self._chooseDimensionWidget.dimension[0]
        except IndexError:
            return 0

    @property
    def dimension2(self) -> int:
        try:
            return self._chooseDimensionWidget.dimension[1]
        except IndexError:
            return 1

    def getMosaicity(self):
        return self._mosaicity

    def getContoursImage(self):
        return self._contoursPlot.getImage()

    def getOrientationDist(self):
        return self._orientation_dist_data

    def setDataset(self, dataset: ImageDataset):
        self._dataset = dataset

        with block_signals(self._chooseDimensionWidget):
            self._chooseDimensionWidget.setDimensions(self._dataset.dims)
            self._chooseDimensionWidget._updateState(True)

        self._contoursPlot.setGraphTitle(
            self._dataset.title + "\n" + MultiDimMomentType.ORIENTATION_DIST.value
        )
        self._mosaicityPlot.setKeepDataAspectRatio(True)
        self._mosaicityPlot.setGraphTitle(
            self._dataset.title + "\n" + MapType.MOSAICITY.value
        )

        self._computeMosaicityAndOriDist()

    def _plotMosaicity(self):
        if self._dataset is None:
            return

        if self._mosaicity is not None:
            add_image_with_transformation(
                self._mosaicityPlot, self._mosaicity, self._dataset.transformation
            )

    def _plotOrientationData(self, state=None):
        if self._dataset is None or self._orientation_dist_data is None:
            return

        self._contoursPlot.remove(kind="curve")
        self._contoursPlot.resetZoom()

        origin = self._orientation_dist_data.origin(
            self._axisTypeCB.getCurrentAxisType()
        )

        if self._oriDistTypeButtons.checkedId() == OriDistButtonIds.COLOR_KEY.value:
            data = self._orientation_dist_data.rgb_key
            scale = self._orientation_dist_data.rgb_key_plot_scale()
        else:
            data = self._orientation_dist_data.data
            scale = self._orientation_dist_data.data_plot_scale()

        self._contoursPlot.addImage(
            data,
            xlabel=self._orientation_dist_data.x_label,
            ylabel=self._orientation_dist_data.y_label,
            origin=origin,
            scale=scale,
        )
        self._computeContours()

    def _computeContours(self):
        """
        Compute contours map based on orientation distribution.
        """
        self._contoursPlot.remove(kind="curve")
        orientation_image_plot: ImageBase | None = self._contoursPlot.getImage()

        if self._orientation_dist_data is None or orientation_image_plot is None:
            return

        min_orientation = numpy.min(self._orientation_dist_data.data)
        max_orientation = numpy.max(self._orientation_dist_data.data)

        polygons = []
        levels = []
        for i in numpy.linspace(
            min_orientation, max_orientation, int(self._levelsLE.text())
        ):
            polygons.append(find_contours(self._orientation_dist_data.data, i))
            levels.append(i)

        colormap = Colormap(
            name="temperature", vmin=min_orientation, vmax=max_orientation
        )
        colors = colormap.applyToData(levels)
        self.contours = {}
        for ipolygon, polygon in enumerate(polygons):
            # iso contours
            for icontour, contour in enumerate(polygon):
                if len(contour) == 0:
                    continue

                x = contour[:, 1]
                y = contour[:, 0]
                x_pts, y_pts = self._orientation_dist_data.to_motor_coordinates(
                    x, y, self._axisTypeCB.getCurrentAxisType()
                )
                legend = f"poly{icontour}.{ipolygon}"
                self.contours[legend] = {
                    "points": (x_pts.copy(), y_pts.copy()),
                    "color": colors[ipolygon],
                    "value": levels[ipolygon],
                    "pixels": (x, y),
                }
                self._contoursPlot.addCurve(
                    x=x_pts,
                    y=y_pts,
                    linestyle="-",
                    linewidth=2.0,
                    legend=legend,
                    resetzoom=False,
                    color=colors[ipolygon],
                )

    def _computeMosaicityAndOriDist(self):
        """
        Compute mosaicity and orientation distribution.

        Called when dimensions are changed (by setting the dataset or user interaction).
        """

        if self._dataset is None:
            return

        self._mosaicity = compute_mosaicity(
            self._dataset.moments_dims,
            x_dimension=self.dimension1,
            y_dimension=self.dimension2,
        )

        self._orientation_dist_data = OrientationDistData(
            self._dataset,
            x_dimension=self.dimension1,
            y_dimension=self.dimension2,
        )
        self._chooseDimensionWidget.setVisible(self._dataset.dims.ndim == 3)
        self._plotMosaicity()
        self._plotOrientationData()



from __future__ import annotations

import logging
from contextlib import contextmanager
from enum import Enum
from typing import Generator

import h5py
from silx.io.utils import h5py_read_dataset

from darfix.core.dataset import ImageDataset
from darfix.core.dimension import Dimension

_logger = logging.getLogger(__file__)

FSCAN_PARAMETERS = "instrument/fscan_parameters/"

FAST_MOTOR_MODE = "fast_motor_mode"

MOTOR_NAME = "motor"
MOTOR_NPOINTS = "npoints"
MOTOR_START_POS = "start_pos"
MOTOR_STEP_SIZE = "step_size"

SCAN_NAME = "scan_name"


class FScanMotorPrefix(Enum):
    PREFIX_2D_FAST = "fast_"
    PREFIX_2D_SLOW = "slow_"
    PREFIX_1D = ""


class OpenFscanError(Exception):
    def __init__(self, details: str):
        super().__init__(details)


@contextmanager
def _open_fscan_parameters(dataset: ImageDataset) -> Generator[h5py.File, None, None]:
    if dataset.metadata_url is None:
        raise OpenFscanError("Dataset has no metadata.")
    with h5py.File(dataset.metadata_url.file_path()) as h5file:
        scan_number = dataset.metadata_url.data_path().lstrip("/").split("/")[0]
        fscan_parameters = h5file.get(scan_number + "/" + FSCAN_PARAMETERS, None)
        if fscan_parameters is None or not isinstance(fscan_parameters, h5py.Group):
            raise OpenFscanError("fscan parameters group not found.")

        yield fscan_parameters


def fscan_get_dimensions(
    dataset: ImageDataset,
) -> tuple[bool, dict[int, Dimension]] | None:
    """
    :return fscan_parameters:  a tuple (is_zigzag, dimensions) or None if no fscan_parameters
    """
    try:
        with _open_fscan_parameters(dataset) as fscan_parameters:

            is_zig_zag = False

            fast_motor_mode = fscan_parameters.get(FAST_MOTOR_MODE, None)
            if fast_motor_mode is not None:
                is_zig_zag = h5py_read_dataset(fast_motor_mode) == "ZIGZAG"
            else:
                is_zig_zag = False
            if h5py_read_dataset(fscan_parameters[SCAN_NAME]) == "fscan":
                # 1D scan
                motor = _get_motor_dimension(
                    fscan_parameters, FScanMotorPrefix.PREFIX_1D
                )
                return is_zig_zag, {0: motor}

            if h5py_read_dataset(fscan_parameters[SCAN_NAME]) == "fscan2d":
                # 2D scan
                fast_motor = _get_motor_dimension(
                    fscan_parameters, FScanMotorPrefix.PREFIX_2D_FAST
                )
                slow_motor = _get_motor_dimension(
                    fscan_parameters, FScanMotorPrefix.PREFIX_2D_SLOW
                )
                return is_zig_zag, {0: fast_motor, 1: slow_motor}

    except OpenFscanError as e:
        _logger.debug(e)

    return None


def _get_motor_dimension(
    parameters: h5py.Group, motor_prefix: FScanMotorPrefix
) -> Dimension:
    """
    Create Dimension object from fscan_parameter for the given motor prefix
    """
    name = h5py_read_dataset(parameters[motor_prefix.value + MOTOR_NAME])
    npoints = h5py_read_dataset(parameters[motor_prefix.value + MOTOR_NPOINTS])
    start_pos = h5py_read_dataset(parameters[motor_prefix.value + MOTOR_START_POS])
    step_size = h5py_read_dataset(parameters[motor_prefix.value + MOTOR_STEP_SIZE])

    return Dimension(
        name,
        npoints,
        start_pos,
        start_pos + step_size * npoints,
    )



__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "02/09/2019"


import numpy


def normalized_variance(img):
    """
    Computes the normalized variance autofocus function into the given image.
    """
    img = numpy.asanyarray(img)
    with numpy.errstate(invalid="ignore", divide="ignore"):
        return img.var() / img.mean()



import os

import h5py
import numpy
import pytest
from silx.io.dictdump import dicttoh5

from darfix import dtypes
from darfix.tasks.hdf5_data_selection import HDF5DataSelection


@pytest.mark.parametrize("provide_metadata", (True, False))
@pytest.mark.parametrize("provide_bg", (True, False))
def test_hdf5_data_selection(tmp_path, provide_metadata: bool, provide_bg: bool):
    """test load_process_data function with HDF5 dataset"""
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()

    raw_data_file = os.path.join(raw_data_dir, "raw.hdf5")
    with h5py.File(raw_data_file, mode="w") as h5f:
        h5f["/path/to/data"] = numpy.arange(0, 100 * 100 * 20).reshape(20, 100, 100)

    if provide_metadata:
        dicttoh5(
            {
                "positioners": {
                    "alpha": 1,
                    "beta": numpy.linspace(0, 1, num=20),
                }
            },
            h5file=raw_data_file,
            h5path="/path/instrument",
            mode="a",
        )
        raw_metadata_path = "/path/instrument/positioners"
    else:
        raw_metadata_path = None

    if provide_bg:
        raw_dark_file = os.path.join(raw_data_dir, "dark.hdf5")
        with h5py.File(raw_dark_file, mode="w") as h5f:
            h5f["/path/to/dark"] = numpy.arange(100 * 100).reshape(1, 100, 100)
        bg_detector_data_path = "/path/to/dark"
    else:
        raw_dark_file = None
        bg_detector_data_path = None

    task = HDF5DataSelection(
        inputs={
            "raw_input_file": raw_data_file,
            "raw_detector_data_path": "/path/to/data",
            "dark_input_file": raw_dark_file,
            "dark_detector_data_path": bg_detector_data_path,
            "raw_metadata_path": raw_metadata_path,
        }
    )
    task.run()
    assert isinstance(task.outputs.dataset, dtypes.Dataset)
    assert task.outputs.dataset.dataset.data is not None
    if provide_bg:
        assert task.outputs.dataset.bg_dataset is not None
    else:
        assert task.outputs.dataset.bg_dataset is None

    if provide_metadata:
        numpy.testing.assert_array_equal(
            task.outputs.dataset.dataset.get_metadata_values(key="beta"),
            numpy.linspace(0, 1, num=20),
        )
    else:
        numpy.testing.assert_array_equal(
            task.outputs.dataset.dataset.get_metadata_values(key="beta"),
            numpy.array([numpy.nan] * 20),
        )



import functools
from typing import Optional

from ewokscore.missing_data import MISSING_DATA
from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread

from darfix import dtypes
from darfix.gui.dataPartitionWidget import DataPartitionWidget
from darfix.gui.operationThread import OperationThread
from darfix.gui.utils.message import missing_dataset_msg
from darfix.tasks.datapartition import DataPartition


class DataPartitionWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=DataPartition):
    """
    In cases with a large number of images you may want to omit the images with low intensity.
    This widget allows you to see an intensity curve of the images and to choose how many images you want to keep.
    At the next steps of the workflow only the images with higher intensity will be used for the analysis.
    """

    name = "partition data"
    icon = "icons/filter.png"
    want_main_area = False

    _ewoks_inputs_to_hide_from_orange = (
        "bins",
        "filter_bottom_bin_idx",
        "filter_top_bin_idx",
    )

    def __init__(self):
        super().__init__()
        self._threadComputeHistogram = None

        self._widget = DataPartitionWidget(parent=self)
        self.controlArea.layout().addWidget(self._widget)

        # update gui from default input value
        bins = self.get_default_input_value("bins", None)
        if bins is not None:
            self._widget.setBins(bins)

        filter_bottom_bin_idx = self.get_default_input_value(
            "filter_bottom_bin_idx", None
        )
        if filter_bottom_bin_idx is not None:
            self._widget.setBottomBin(filter_bottom_bin_idx)

        filter_top_bin_idx = self.get_default_input_value("filter_top_bin_idx", None)
        if filter_top_bin_idx is not None:
            self._widget.setTopBin(filter_top_bin_idx)

        # connect signal / slot
        self._widget.sigAbort.connect(self._abort_processing)
        self._widget.sigComputeHistogram.connect(self._computeHistogram)
        self._widget.sigPartitionData.connect(self.execute_ewoks_task)
        # keep ewoks default input value up to date.
        self._widget.sigNbBinsChanged.connect(self._nbBinsHasChanged)
        self._widget.sigBottomBinChanged.connect(self._bottomBinHasChanged)
        self._widget.sigTopBinChanged.connect(self._topBinHasChanged)

    def setDataset(self, dataset: Optional[dtypes.Dataset], pop_up=False):
        if dataset is None:
            return
        if not isinstance(dataset, dtypes.Dataset):
            raise dtypes.DatasetTypeError(dataset)
        self.set_dynamic_input("dataset", dataset)

        self._widget.setDataset(dataset)
        if pop_up:
            self.open()

    def handleNewSignals(self) -> None:
        """
        Today the DataPartitionWidget is not processing automatically a dataset when it receive it.
        It wait the user to press 'filter' in this case to process and move to the next widget.
        """
        dataset = self.get_task_input_value("dataset", MISSING_DATA)
        if dataset not in (MISSING_DATA, None):
            self.setDataset(dataset=dataset, pop_up=True)
        # note: we only want to execute the ewows task when 'ok' is pressed and not when a new dataset is set
        # return super().handleNewSignals() do not call to make sure the processing is not triggered

    def _computeHistogram(self):
        """
        compute the histogram and display it.
        It will help user select the bottom and top bins values.
        """
        dataset = self.get_task_input_value("dataset", MISSING_DATA)
        if dataset in (None, MISSING_DATA):
            missing_dataset_msg()
            return
        if not isinstance(dataset, dtypes.Dataset):
            raise dtypes.DatasetTypeError(dataset)
        self._threadComputeHistogram = OperationThread(
            self, dataset.dataset.compute_frames_intensity
        )
        self._widget.setProcessingButtonsEnabled(False)
        self._threadComputeHistogram.finished.connect(
            self._widget._showHistogramCallback
        )
        self._threadComputeHistogram.finished.connect(
            functools.partial(
                self._threadComputeHistogram.finished.disconnect,
                self._widget._showHistogramCallback,
            )
        )
        self._threadComputeHistogram.start()

    def _abort_processing(self):
        "stop all processing associated to the task or the widget"
        self.cancel_running_task()
        if self._threadComputeHistogram and self._threadComputeHistogram.isRunning():
            self._threadComputeHistogram.quit()

    def _nbBinsHasChanged(self, nb_bins):
        self.set_default_input("bins", nb_bins)

    def _bottomBinHasChanged(self, value):
        self.set_default_input("filter_bottom_bin_idx", value)

    def _topBinHasChanged(self, value):
        self.set_default_input("filter_top_bin_idx", value)



import logging

from silx.gui import qt

from ..utils.utils import create_completer

_logger = logging.getLogger(__name__)


class WorkingDirSelectionWidget(qt.QWidget):
    """
    Widget used to obtain a directory name
    """

    dirChanged = qt.Signal()

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)

        self._dir = qt.QLineEdit("", parent=self)
        self._dir.editingFinished.connect(self.dirChanged)
        self.completer = create_completer()
        self._dir.setCompleter(self.completer)

        self._addButton = qt.QPushButton("Select working directory", parent=self)
        self._addButton.pressed.connect(self._selectWorkingDirectory)
        self.setLayout(qt.QHBoxLayout())

        self.layout().addWidget(self._dir)
        self.layout().addWidget(self._addButton)

    def _selectWorkingDirectory(self):
        """
        Select a folder to be used as working directory (task results will be saved at this location)
        """
        fileDialog = qt.QFileDialog()
        fileDialog.setOption(qt.QFileDialog.ShowDirsOnly)
        fileDialog.setFileMode(qt.QFileDialog.Directory)
        if fileDialog.exec():
            self._dir.setText(fileDialog.directory().absolutePath())
            self.dirChanged.emit()
        else:
            _logger.warning("Could not open directory")

    def getDir(self):
        return str(self._dir.text())

    def setDir(self, _dir):
        self._dir.setText(str(_dir))
        self.dirChanged.emit()



from __future__ import annotations

from silx.gui import qt

from .WorkingDirSelectionWidget import WorkingDirSelectionWidget


class DataSelectionBase(qt.QTabWidget):
    """Define the base class to define the data selection"""

    sigRawDataInfosChanged = qt.Signal()
    sigDarkDataInfosChanged = qt.Signal()
    sigTreatedDirInfoChanged = qt.Signal()

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)

        self._dataset = None
        self.bg_dataset = None
        self.indices = None
        self.bg_indices = None

        self._rawDataWidget = self.buildRawDataWidget()
        self.addTab(self._rawDataWidget, "raw data")

        self._darkDataWidget = self.buildDarkDataWidget()
        self.addTab(self._darkDataWidget, "dark data")

        self._treatedDirWidget = self.buildProcessedDataWidget()
        self.addTab(self._treatedDirWidget, "treated data")

        # connect signal / slot
        self._treatedDirData.dirChanged.connect(self.sigTreatedDirInfoChanged)

    def getTreatedDir(self) -> str:
        return self._treatedDirData.getDir()

    def setTreatedDir(self, _dir):
        self._treatedDirData.setDir(_dir)

    def buildProcessedDataWidget(self):
        widget = qt.QWidget()
        widget.setLayout(qt.QVBoxLayout())
        self._treatedDirData = WorkingDirSelectionWidget(parent=self)
        widget.layout().addWidget(self._treatedDirData)
        spacer = qt.QWidget()
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        widget.layout().addWidget(spacer)
        return widget

    # Base class API to be redefined
    def buildRawDataWidget(self):
        raise NotImplementedError("Base class")

    def buildDarkDataWidget(self):
        raise NotImplementedError("Base class")

    def keepDataOnDisk(self) -> bool:
        raise NotImplementedError("Base class")



from __future__ import annotations

import logging
import os
from typing import Optional

import h5py
import numpy
from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.missing_data import is_missing_data
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict
from pydantic import Field
from silx.io.dictdump import dicttoh5
from silx.utils.enum import Enum as _Enum

from darfix.core.datapathfinder import DETECTOR_KEYWORD
from darfix.core.datapathfinder import SCAN_KEYWORD
from darfix.core.datapathfinder import DataPathFinder
from darfix.core.datapathfinder import sort_bliss_scan_entries
from darfix.core.settings import PROCESSED_DATA
from darfix.core.settings import RAW_DATA

_logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class VDS_POLICY(_Enum):
    """
    Virtual dataset policy. Values can be:
    * 'absolute': in this case the links will be done with the absolute path. Safer for 'single' shot processing.
    * 'relative': in this case the links will be done with relative path. Safer if we want to move files. Links will be preserved as long as relative paths are preserved.
    """

    RELATIVE = "relative"
    ABSOLUTE = "absolute"


def _concatenate_dict(
    dict_1: dict[str, numpy.ndarray], dict_2: dict[str, numpy.ndarray]
) -> dict[str, numpy.ndarray]:
    """
    concatenate two dicts into a third dict. Keys are str and values are numpy.ndarray.
    If a dict contains the same key then the two values (numpy array) will be concatenated.
    """
    assert isinstance(dict_1, dict)
    assert isinstance(dict_2, dict)
    res = {}
    # concatenate keys. Note: we create a new list to keep keys ordering and be consistent.
    # creating a set for example we reorder the keys
    keys = list(dict_1.keys())
    keys.extend(filter(lambda key: key not in dict_1.keys(), dict_2.keys()))
    for key in keys:
        if key in dict_1.keys() and key in dict_2.keys():
            res[key] = numpy.concatenate((dict_1[key], dict_2[key]))
        elif key in dict_1.keys():
            res[key] = dict_1[key]
        else:
            res[key] = dict_2[key]
    return res


def _check_positioners_consistency(my_dict: dict) -> None:
    """
    make sure all the values of 'my_dict' have the same number of elements
    """
    n_elmts = numpy.median([len(value) for value in my_dict.values()])
    for key, value in my_dict.items():
        if len(value) != n_elmts:
            _logger.warning(
                f"Found inconsistent positioner dataset '{key}'. Get {len(value)} elements when {n_elmts} expected"
            )


def _filter_static_positioners(my_dict: dict) -> None:
    """
    replace all positioners which have a unique value by a scalar
    """
    keys = tuple(my_dict.keys())
    for key in keys:
        uniques = numpy.unique(my_dict[key])
        if len(uniques) == 1:
            my_dict[key] = uniques[0]


def create_virtual_source(
    input_dataset: h5py.Dataset,
    output_file: str,
    vds_policy: str | VDS_POLICY = VDS_POLICY.RELATIVE,
) -> h5py.VirtualSource:
    """
    create the VirtualSource according to the defined policy
    """
    vds_policy = VDS_POLICY(vds_policy)
    if vds_policy is VDS_POLICY.ABSOLUTE:
        return h5py.VirtualSource(input_dataset)
    elif vds_policy is VDS_POLICY.RELATIVE:
        relpath = os.path.relpath(
            os.path.abspath(input_dataset.file.filename),
            os.path.dirname(os.path.abspath(output_file)),
        )
        if not relpath.startswith("./"):
            relpath = "./" + relpath
        return h5py.VirtualSource(
            path_or_dataset=relpath,
            name=input_dataset.name,
            shape=input_dataset.shape,
            dtype=input_dataset.dtype,
        )
    else:
        raise ValueError(
            f"VDS_POLICY should be 'absolute' or 'relative'. Get '{VDS_POLICY}'"
        )


def concatenate_scans(
    input_file: str,
    entries_to_concatenate: Optional[tuple],
    output_file: str,
    detector_data_path: str,
    positioners_group_path: str,
    output_entry_name: str = "entry_0000",
    overwrite: bool = False,
    vds_policy: str | VDS_POLICY = VDS_POLICY.RELATIVE,
) -> None:
    """
    :param input_file: proposal file containing link to all the detector frame...
    :param entries_to_concatenate: tuple of all entries to concatenate. Order will be preserved. If None provided then all entries will be concatenated
    :param output_file: location of the output file
    :param detector_data_path: path to the detector dataset. Expected to be provided as '{scan}/path/to/detector/dataset' (a) or '{scan}/path/to/{detector}' (b).

        * in the use case (a) the relative (to {scan}) path is fully provided and will be take 'as such'
        * in the use case (b) only a folder path is provided and the keyword {detector} must be provided at the end. In this case the function will call the 'find_detector_dataset' function. And will browse the group for any detector.
          First groups with 'nexus' attributes fitting a 3D detector will be search. If none are found then it will return the first 3D dataset found.
    :param positioners_group_path: path to the positioners datasets (containing motor positions). Expected to be provided as a '{scan}/path/to/group' pattern. Where the '{scan}' part will be replaced by input file first level items.
    :param output_entry_name: HDF5 group name that will contain the concatenated detector data + metadata.
    :param overwrite: if False and output file exist then will not overwrite it
    :param vds_policy: policy regarding the VirtualDataSet. Either "absolute" or "relative"
    """
    _logger.info("start concatenation")
    # check inputs
    if os.path.exists(output_file) and not overwrite:
        raise OSError(
            f"output file exists ({output_file}). Please remove it before processing or set 'overwrite to True'"
        )
    if not isinstance(detector_data_path, str):
        raise TypeError("detector_dataset_path should be a str")
    if not isinstance(output_entry_name, str):
        raise TypeError("output_entry_name should be a str")
    if not isinstance(positioners_group_path, str):
        raise TypeError("positioners_dataset_path should be a str")

    detector_data_path_finder = DataPathFinder(
        file_=input_file,
        pattern=detector_data_path,
        filter_entries=entries_to_concatenate,
        allowed_keywords=(SCAN_KEYWORD, DETECTOR_KEYWORD),
    )
    positioners_data_path_finder = DataPathFinder(
        file_=input_file,
        pattern=positioners_group_path,
        filter_entries=entries_to_concatenate,
        allowed_keywords=(SCAN_KEYWORD, DETECTOR_KEYWORD),
    )

    # concatenate
    with h5py.File(input_file, mode="r") as h5f_input:
        if entries_to_concatenate is None:
            # sort entries to concatenate
            try:
                entries_to_concatenate = sort_bliss_scan_entries(h5f_input.keys())
            except ValueError:
                _logger.error("Failed to order scans by indices. Take them 'unordered'")
                entries_to_concatenate = tuple(h5f_input.keys())

        _logger.info(f"(sorted) entries to concatenate {entries_to_concatenate}")
        entries_n_frame = []
        # store the number of frames along all entries
        frame_shape = None
        detector_data_type = None
        positioners = {}
        virtual_sources = []

        # number of frame concatenateed
        for entry in entries_to_concatenate:

            # update the entry_detector_path. Note: first_scan and last_scan have no meaning in the case
            # of concatenation.
            entry_detector_path = detector_data_path_finder.format(
                scan=entry, first_scan=None, last_scan=None
            )

            if entry_detector_path not in h5f_input:
                _logger.error(
                    f"Unable to find detector path '{entry_detector_path}' from file '{input_file}'"
                )
                continue
            if h5f_input[entry_detector_path].ndim != 3:
                raise ValueError(
                    f"detector dataset are expected to be 3D. Get {h5f_input[entry_detector_path].ndim}"
                )
            # 1.1: get metadata from the dataset and make sure it is coherent along all the detector datasets
            entry_n_frame = h5f_input[entry_detector_path].shape[0]
            entry_frame_shape = h5f_input[entry_detector_path].shape[1:]
            entries_n_frame.append(entry_n_frame)
            if frame_shape is None:
                frame_shape = entry_frame_shape
            elif entry_frame_shape != frame_shape:
                raise ValueError(
                    f"Incoherent frame shape. {entry} get {entry_frame_shape} when {frame_shape} expected"
                )
            if detector_data_type is None:
                detector_data_type = h5f_input[entry_detector_path].dtype
            elif detector_data_type != h5f_input[entry_detector_path].dtype:
                raise TypeError(
                    f"Inconsistent data type between the scan. {entry} get {h5f_input[entry_detector_path].dtype} when {detector_data_type} expected"
                )
            # 1.2: create VirtualSource to be used once entries browse

            virtual_sources.append(
                create_virtual_source(
                    input_dataset=h5f_input[entry_detector_path],
                    output_file=output_file,
                    vds_policy=vds_policy,
                )
            )
            # 2.0 handle positioners.
            # note: positioners dataset will be copied
            # number of frame in the current entry / scan
            entry_positioner_path = positioners_data_path_finder.format(
                scan=entry, first_scan=None, last_scan=None
            )
            if entry_positioner_path not in h5f_input:
                _logger.error(
                    f"Unable to find positioners path '{entry_positioner_path}' from file '{input_file}'"
                )
                continue
            entry_positioners_grp = h5f_input[entry_positioner_path]
            # HDF5 group containing the positioners
            entry_positioners = {}
            # dict used to concatenate all the positioners as numpy array
            for key in entry_positioners_grp:
                dataset = entry_positioners_grp[key]
                if not isinstance(dataset, h5py.Dataset):
                    _logger.warning(
                        f"Found a none h5py.Dataset in entry {entry_positioners_grp}: {key}"
                    )
                elif dataset.ndim > 2:
                    # in case a dataset with more than 2 dimensions is part of the positioners, simply ignore it.
                    _logger.debug(f"Skip a dataset higher than 2D ({dataset.name})")
                else:
                    value = dataset[()]
                    if numpy.isscalar(value) or len(value.shape) == 0:
                        # convert scalars to arrays. As a value can be static in a scan context but
                        # dynamic / array in the scope of the full acquisition
                        value = numpy.array([value] * entry_n_frame)
                    entry_positioners[key] = value

            positioners = _concatenate_dict(positioners, entry_positioners)

        # create output directory in case not existing (h5py won't do it)
        os.makedirs(
            os.path.dirname(os.path.abspath(output_file)),
            exist_ok=True,
        )
        # write the detector virtual dataset (VDS) to output file
        with h5py.File(output_file, mode="a") as h5f_output:
            raw_entry_name = detector_data_path.split("/")[0]
            entry_name = raw_entry_name.format(
                scan=output_entry_name,
                first_scan=None,
                last_scan=None,
            )
            if entry_name in h5f_output and overwrite:
                del h5f_output[entry_name]

            output_detector_dataset_path = detector_data_path_finder.format(
                scan=output_entry_name,
                first_scan=None,
                last_scan=None,
            )
            n_frames = numpy.sum(entries_n_frame)
            virtual_layout = h5py.VirtualLayout(
                shape=(n_frames, frame_shape[0], frame_shape[1]),
                dtype=detector_data_type,
            )
            assert len(virtual_sources) == len(
                entries_n_frame
            ), "we expect one virtual source per entry"
            virtual_layout_index = 0
            for entry_n_frame, virtual_source in zip(entries_n_frame, virtual_sources):
                virtual_layout[
                    virtual_layout_index : virtual_layout_index + entry_n_frame
                ] = virtual_source
                virtual_layout_index += entry_n_frame

            h5f_output.create_virtual_dataset(
                output_detector_dataset_path, virtual_layout
            )
    # check number of elements is coherent else warm the user
    _check_positioners_consistency(positioners)
    _filter_static_positioners(positioners)

    dicttoh5(
        positioners,
        h5file=output_file,
        h5path=positioners_group_path.format(scan=output_entry_name),
        mode="a",
    )

    _logger.info(
        f"concatenation finished. You can run 'silx view {output_file}' to check the result"
    )


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    input_file: str = Field(
        examples=["/path/to/input/file.h5"],
        description="Path to the input file containing scans to concatenate.",
    )
    entries_to_concatenate: tuple[str, ...] | MissingData = Field(
        default=MISSING_DATA,
        examples=[("/1.1", "/2.1", "/3.1")],
        description="Entries (scans) in the file to concatenate. If not provided, all entries will be concatenated.",
    )
    detector_data_path: str | MissingData = Field(
        default=MISSING_DATA,
        examples=[
            "{scan}/instrument/measurement/my_detector",
            "{scan}/instrument/measurement/{detector}",
        ],
        description="Path pattern to the detector data in the input file. If `{detector}` is in the pattern then all the datasets from the subpath will be browsed in order to 'guess' the detector to be used. If not provided, {scan}/measurement/{detector} is the default pattern",
    )
    positioners_group_path: str | MissingData = Field(
        default=MISSING_DATA,
        examples=["{scan}/instrument/positioners"],
        description="Path pattern to the positioners group in the input file. If not provided, {scan}/instrument/positioners is the default pattern",
    )
    overwrite: bool | MissingData = Field(
        default=MISSING_DATA,
        description="If True then will overwrite the output file if it exists. False if not provided.",
    )
    output_file: str | MissingData = Field(
        default=MISSING_DATA,
        examples=["/path/to/output/file.h5"],
        description="Path to the output file. Must be provided if guess_output_file is False.",
    )
    """Path to the output file. If not provided then will try to guess it from the input file."""
    guess_output_file: bool | MissingData = Field(
        default=MISSING_DATA,
        description="If True then will try to guess the output file from the input file. False if not provided.",
    )


class ConcatenateHDF5Scans(
    Task,
    input_model=Inputs,
    output_names=[
        "output_file",
    ],
):
    """
    Concatenate a set of scans / entries contained in 'input_file'.
    If entries_to_concatenate is None then all entries will be concatenated

    * 'detector_data_path' is the pattern to find all the detector data path. It
    is expected to look like '{scan}/instrument/measurement/my_detector'. In this case
    it will look for each scan at the same location. So for the scan '1.1' it
    will look for '1.1/instrument/measurement/my_detector'. For the scan 2.1 it will look for '2.1/instrument/measurement/my_detector'...
    If the {detector} is provided then all the dataset part of the upper path will be browse in order to 'guess' the detector to be used.
    Using the 'find_detector' function.
    * 'positioners_group_path' is the pattern to find all positioner groups. And should look like '{scan}/instrument/positioners'
    For the scan '1.1' it will look for '1.1/instrument/positioners'...
    """

    DEFAULT_DETECTOR_DATA_PATH = SCAN_KEYWORD + "/measurement/" + DETECTOR_KEYWORD

    DEFAULT_POSITIONERS_DATA_PATH = SCAN_KEYWORD + "/instrument/positioners"

    def run(self):
        detector_data_path = self.get_input_value(
            "detector_data_path", self.DEFAULT_DETECTOR_DATA_PATH
        )
        positioners_group_path = self.get_input_value(
            "positioners_group_path", self.DEFAULT_POSITIONERS_DATA_PATH
        )
        output_file = self.get_input_value("output_file")
        if is_missing_data(output_file):
            if not self.get_input_value("guess_output_file", False):
                raise ValueError(
                    "Either the output file should be provided or you should ask to determine automatically output file ('guess_output_file')"
                )
            output_file = guess_output_file(input_file=output_file)
        concatenate_scans(
            input_file=self.inputs.input_file,
            entries_to_concatenate=self.get_input_value("entries_to_concatenate", None),
            output_file=output_file,
            detector_data_path=detector_data_path,
            positioners_group_path=positioners_group_path,
            overwrite=self.get_input_value("overwrite", False),
        )
        self.outputs.output_file = self.inputs.output_file


def find_scan_data_path(file_path, pattern_data_path):
    with h5py.File(file_path) as h5f:
        pattern_data_path = pattern_data_path.lstrip("/")
        if pattern_data_path.startswith("{scan}"):
            try:
                first_scan = h5f.get(next(h5f.keys()))
            except StopIteration:
                # in case the hdf5 is empty
                return None
            path_search = pattern_data_path.replace("{scan}", "")
        else:
            first_scan = None
            path_search = pattern_data_path

        if path_search not in h5f:
            return None

        if first_scan is not None:
            result = pattern_data_path.replace(scan=first_scan.name)
        else:
            result = path_search
        return result


def guess_output_file(input_file: str, target_processed_data_dir: bool = True):
    """
    propose an output file path for scan concatenation from an input file

    :param input_file: file containing the scans to concatenate
    :param target_processed_data_dir: If true then will try to make the output file as part of 'PROCESSED_DATA' folder.
    """

    file_path, ext = os.path.splitext(input_file)
    output_file = "".join(
        (
            "_".join((file_path, "darfix_concat")),
            ext,
        )
    )
    if target_processed_data_dir:
        splitted_paths = output_file.split(os.path.sep)
        # reverse it to find the lower level value of '_RAW_DATA_DIR_NAME' if by any 'chance' has several in the path
        # in this case this is most likely what we want
        splitted_paths = splitted_paths[::-1]
        try:
            index_raw_data = splitted_paths.index(RAW_DATA)
        except ValueError:
            # in the case RAW_DATA dir name is not contained
            pass
        else:
            splitted_paths[index_raw_data] = PROCESSED_DATA
        output_file = os.sep.join(splitted_paths[::-1])
    return output_file



import pytest

from darfix.core.utils import NoDimensionsError


def test_apply_moments(dataset):

    with pytest.raises(NoDimensionsError):
        dataset.apply_moments(indices=[1, 2, 3, 4])

    dataset.find_dimensions()
    dataset.reshape_data()
    moments = dataset.apply_moments(indices=[1, 2, 3, 4])
    assert moments[0][0].shape == dataset.frame_shape
    assert moments[1][3].shape == dataset.frame_shape



from __future__ import annotations

from functools import partial
from typing import Any

from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread
from silx.gui import qt

from darfix.gui.data_selection.scan_selection_widgets import ScanSelectionWidget
from darfix.gui.data_selection.scan_selection_widgets import (
    ScanSelectionWidgetWithPositioners,
)
from darfix.gui.utils.message import show_error_msg
from darfix.tasks.hdf5_data_selection import HDF5DataSelection


class HDF5DataSelectionWidgetOW(
    OWEwoksWidgetOneThread,
    ewokstaskclass=HDF5DataSelection,
):
    """
    Widget to select dataset stored as HDF5
    """

    name = "HDF5 data selection"
    icon = "icons/upload_hdf5.svg"
    want_main_area = True
    want_control_area = False

    priority = 1

    _ewoks_inputs_to_hide_from_orange = (
        "raw_detector_data_path",
        "raw_metadata_path",
        "dark_detector_data_path",
        "workflow_title",
        "treated_data_dir",
    )

    def __init__(self):
        super().__init__()

        layout = self.mainArea.layout()
        rawDataGroup = qt.QGroupBox(title="Raw data")
        rawDataLayout = qt.QVBoxLayout(rawDataGroup)
        formLayout = qt.QFormLayout()
        self._rawDataSelection = ScanSelectionWidgetWithPositioners()

        rawDataLayout.addWidget(self._rawDataSelection)
        rawDataLayout.addLayout(formLayout)
        layout.addWidget(rawDataGroup)

        self._darkDataGroup = qt.QGroupBox(title="Dark data")
        self._darkDataGroup.setCheckable(True)
        self._darkDataGroup.setChecked(False)
        darkDataLayout = qt.QVBoxLayout(self._darkDataGroup)
        self._darkDataSelection = ScanSelectionWidget()
        darkDataLayout.addWidget(self._darkDataSelection)
        layout.addWidget(self._darkDataGroup)

        workingDirLayout = qt.QFormLayout()
        self._workingDirLineEdit = qt.QLineEdit()
        workingDirLayout.addRow("Working directory", self._workingDirLineEdit)
        self._workflowTitleLineEdit = qt.QLineEdit()
        workingDirLayout.addRow("Workflow title", self._workflowTitleLineEdit)
        layout.addLayout(workingDirLayout)

        types = qt.QDialogButtonBox.Ok
        _buttons = qt.QDialogButtonBox(parent=self)
        _buttons.setStandardButtons(types)
        self.mainArea.layout().addWidget(_buttons)

        _buttons.accepted.connect(self._executeTask)

        # set up
        self._loadInputs(self.get_task_input_values())

        # connect signal / slot
        self.task_executor.finished.connect(
            self.information,
        )
        self.task_executor.finished.connect(lambda: self.mainArea.setEnabled(True))
        self.task_executor.started.connect(
            partial(self.information, "Downloading dataset. This can take a while...")
        )
        self.task_executor.started.connect(lambda: self.mainArea.setDisabled(True))

    def task_output_changed(self) -> None:
        if self.task_succeeded:
            self.accept()
        else:
            show_error_msg(f"Selection failed with error : \n{self.task_exception}")

        return super().task_output_changed()

    def _loadInputs(self, taskInputValues: dict[str, Any]):
        rawInputFile = taskInputValues.get("raw_input_file")
        if rawInputFile:
            self._rawDataSelection.setFilePath(rawInputFile)

        rawDetectorDataPath = taskInputValues.get("raw_detector_data_path")
        if rawDetectorDataPath:
            self._rawDataSelection.setDetectorPath(rawDetectorDataPath)

        rawMetadataPath = taskInputValues.get("raw_metadata_path")
        if rawMetadataPath:
            self._rawDataSelection.setPositionersPath(rawMetadataPath)

        darkInputFile = taskInputValues.get("dark_input_file")
        if darkInputFile:
            self._darkDataSelection.setFilePath(darkInputFile)
            self._darkDataGroup.setChecked(True)

        darkDetectorDataPath = taskInputValues.get("dark_detector_data_path")
        if darkDetectorDataPath:
            self._darkDataSelection.setDetectorPath(darkDetectorDataPath)
            self._darkDataGroup.setChecked(True)

        treatedDataDir = taskInputValues.get("treated_data_dir")
        if treatedDataDir is not None:
            self._workingDirLineEdit.setText(treatedDataDir)

        workflowTitle = taskInputValues.get("workflow_title")
        if workflowTitle is not None:
            self._workflowTitleLineEdit.setText(workflowTitle)

    def getInputs(self) -> dict[str, Any]:
        inputs = {
            "raw_input_file": self._rawDataSelection.getFilePath(),
            "raw_detector_data_path": self._rawDataSelection.getDetectorPath(),
            "raw_metadata_path": self._rawDataSelection.getPositionersPath(),
        }

        if self._darkDataGroup.isChecked():
            inputs = {
                **inputs,
                "dark_input_file": self._darkDataSelection.getFilePath(),
                "dark_detector_data_path": self._darkDataSelection.getDetectorPath(),
            }

        if self._workingDirLineEdit.text():
            inputs = {
                **inputs,
                "treated_data_dir": self._workingDirLineEdit.text(),
            }

        if self._workflowTitleLineEdit.text():
            inputs = {
                **inputs,
                "workflow_title": self._workflowTitleLineEdit.text(),
            }

        return inputs

    def handleNewSignals(self) -> None:
        # update the input file in case they are provided by another widget (like the hdf5 scan concatenation)
        self._loadInputs(self.get_task_input_values())

    def saveInputs(self):
        for key, value in self.getInputs().items():
            self.set_default_input(key, value)

    def _executeTask(self):
        self.saveInputs()
        self.execute_ewoks_task()

    def closeEvent(self, *a, **kw):
        self.saveInputs()
        super().closeEvent(*a, **kw)



from __future__ import annotations

import numpy
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D

import darfix

from ..core.transformation import Transformation


class WeakBeamWidget(qt.QMainWindow):
    """
    Widget to recover weak beam to obtain dislocations.
    """

    sigValidate = qt.Signal()
    """Emit when user validate weak beam (ok pressed)"""
    sigApplyThreshold = qt.Signal()
    """Emit when user ask to apply a threshold"""
    sigNValueChanged = qt.Signal()
    """Emit when N value has changed"""

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        widget = qt.QWidget()
        layout = qt.QGridLayout()

        self._nLE = qt.QLineEdit("")
        validator = qt.QDoubleValidator()
        self._nLE.setValidator(validator)
        _buttons = qt.QDialogButtonBox(parent=self)
        self._okB = _buttons.addButton(_buttons.Ok)
        self._applyThresholdB = _buttons.addButton(_buttons.Apply)

        self._plot = Plot2D()
        self._plot.setDefaultColormap(
            Colormap(
                name=darfix.config.DEFAULT_COLORMAP_NAME,
                normalization=darfix.config.DEFAULT_COLORMAP_NORM,
            )
        )
        layout.addWidget(
            qt.QLabel("Increase/decrease threshold std by a value of : "), 0, 0
        )
        layout.addWidget(self._nLE, 0, 1)
        layout.addWidget(self._plot, 1, 0, 1, 2)
        layout.addWidget(_buttons, 2, 0, 1, 2)
        widget.setLayout(layout)
        self.setCentralWidget(widget)

        # connect signal / slot
        self._applyThresholdB.clicked.connect(self.sigApplyThreshold)
        self._okB.clicked.connect(self.sigValidate)
        self._nLE.editingFinished.connect(self.sigNValueChanged)

        # set up
        self.nvalue = 1

    @property
    def nvalue(self) -> float:
        text = self._nLE.text()
        value, ok = self.locale().toFloat(text)
        if not ok:
            raise ValueError(f"Could not convert string to float: '{text}'")
        return value

    @nvalue.setter
    def nvalue(self, nvalue: str | int | float):
        toStr = self.locale().toString
        self._nLE.setText(toStr(nvalue))
        self.sigNValueChanged.emit()

    def setResult(
        self, center_of_mass: numpy.ndarray, transformation: None | Transformation
    ):
        self._plot.clear()
        if transformation is None:
            self._plot.addImage(center_of_mass, xlabel="pixels", ylabel="pixels")
        else:
            if self._dataset.transformation.rotate:
                center_of_mass = numpy.rot90(center_of_mass, 3)
            self._plot.addImage(
                center_of_mass,
                origin=transformation.origin,
                scale=transformation.scale,
                xlabel=transformation.label,
                ylabel=transformation.label,
            )

    def setProcessingButtonsEnabled(self, enabled):
        self._applyThresholdB.setEnabled(enabled)
        self._okB.setEnabled(enabled)



import numpy

from darfix.core.imageStack import FixedDimension


def test_zsum(dataset):
    indices = [1, 2, 3, 6]

    dataset.find_dimensions()
    dataset.reshape_data()
    result = numpy.sum(
        dataset.get_filtered_data(
            fixed_dimension=FixedDimension(0, 1), indices=indices
        ),
        axis=0,
    )
    zsum = dataset.zsum(indices=indices, dimension=FixedDimension(0, 1))
    numpy.testing.assert_array_equal(zsum, result)



from __future__ import annotations

import functools
import os.path
from contextlib import AbstractContextManager
from typing import Union

import h5py
from silx.io import utils
from silx.io.url import DataUrl


def is_hdf5(url: Union[DataUrl, str]) -> bool:
    if isinstance(url, DataUrl):
        file_path = url.file_path()
    else:
        try:
            data_url = DataUrl(path=url)
        except Exception:
            file_path = url
        else:
            file_path = data_url.file_path()

    if not file_path or not os.path.exists(file_path):
        raise FileNotFoundError(f"Could not find {file_path}")

    return h5py.is_hdf5(file_path)


class hdf5_file_cache(AbstractContextManager):
    """
    Context manager that keep the last (HDF5) file used open and keeps a pointer to the last dataset read.

    The goal it to speed up reading from hdf5 dataset. As we expect in most of the case to have a single file.

    File will be closed when leaving the context manager.
    Before this PR most of the time was spend on h5py.__getitem__. This function includes:

    solving dataset path, external links (VDS).
    Keeping a pointer to the latest dataset will avoid doing the data path resolution but more importantly
    when it reads an external link or a VDS it will keep the virtual source origin file open.
    And avoid opening and closing the file.

    For safety all this mechanism is done in a context manager to make sure the files ares properly closed
    after processing or if any exception occurs during reading.

    Note: we expect this class to be removed once the io mechanism will be reworked.
    """

    def __enter__(self):
        self.__h5_file = None
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close_h5_file()

    def close_h5_file(self):
        if self.__h5_file:
            self.__h5_file.close()
        self.__h5_file = None

    @functools.lru_cache(maxsize=1)
    def open_file(self, file_path: str):
        """cache the latest HDF5 file open"""
        self.close_h5_file()
        self.__h5_file = h5py.File(file_path, mode="r")
        return self.__h5_file

    @functools.lru_cache(maxsize=1)
    def _get_h5py_object(
        self, file_path: str, data_path: str
    ) -> h5py.Group | h5py.Dataset:
        """
        Return the h5py Object (Group or Dataset) at the given position.
        """
        h5f = self.open_file(file_path)
        if data_path not in h5f:
            raise KeyError(f"Data path from URL '{data_path}' not found")
        return h5f[data_path]

    def get_data(self, url: DataUrl):
        """
        Retrieve data contained in 'url'.
        """
        if url.scheme() == "silx":

            data = self._get_h5py_object(
                file_path=url.file_path(),
                data_path=url.data_path(),
            )

            if not utils.is_dataset(data):
                raise ValueError(f"Data path from URL '{url.path()}' is not a dataset")

            data_slice = url.data_slice()
            if data_slice is not None:
                return utils.h5py_read_dataset(data, index=data_slice)
            else:
                # works for scalar and array
                return utils.h5py_read_dataset(data)
        else:
            return utils.get_data(url=url)



__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "22/12/2020"

import numpy
from silx.gui import qt

from darfix import dtypes
from darfix.tasks.blindsourceseparation import Method

from .displayComponentsWidget import DisplayComponentsWidget
from .operationThread import OperationThread


class BSSWidget(qt.QMainWindow):
    """
    Widget to apply blind source separation.
    """

    sigComputed = qt.Signal(Method, int)

    sigNbComponentsChanged = qt.Signal(int)
    """emit when the number of component has been changed"""

    sigMethodChanged = qt.Signal(str)
    """emit when the method (PCA, NICA...) changed"""

    DEFAULT_METHOD = Method.PCA.value

    DEFAULT_N_COMP = 1

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        # Method widget
        methodLabel = qt.QLabel("Method: ")
        self.methodCB = qt.QComboBox()
        for member in Method.members():
            self.methodCB.addItem(member.name)
            self.methodCB.setItemData(
                self.methodCB.count() - 1,
                Method.get_description(member),
                qt.Qt.ToolTipRole,
            )
        self.methodCB.setCurrentText(self.DEFAULT_METHOD)
        self.methodCB.currentTextChanged.connect(self.sigMethodChanged)
        # Number of components
        nComponentsLabel = qt.QLabel("Num comp:")
        self.nComponentsLE = qt.QLineEdit(str(self.DEFAULT_N_COMP))
        self.nComponentsLE.setValidator(qt.QIntValidator())
        self.nComponentsLE.editingFinished.connect(self._nbComponentEdited)
        # Compute BSS with the number of components
        self.computeButton = qt.QPushButton("Compute")
        self.computeButton.setEnabled(False)
        # Detect optimal number of components
        self.detectButton = qt.QPushButton("Detect number of components")
        self.detectButton.setEnabled(False)
        self.detectButton.clicked.connect(self._detectComp)

        # Add widgets to layout
        layout = qt.QGridLayout()
        layout.addWidget(methodLabel, 0, 0, 1, 1)
        layout.addWidget(self.methodCB, 0, 1, 1, 1)
        layout.addWidget(nComponentsLabel, 0, 2, 1, 1)
        layout.addWidget(self.nComponentsLE, 0, 3, 1, 1)
        layout.addWidget(self.computeButton, 0, 4, 1, 1)
        layout.addWidget(self.detectButton, 1, 4, 1, 1)
        # Top Widget with the type of bss and the number of components to compute
        top_widget = qt.QWidget(self)
        top_widget.setLayout(layout)

        # Widget to display the components
        self._displayComponentsWidget = DisplayComponentsWidget()
        self._displayComponentsWidget.hide()

        # Main widget is a Splitter with the top widget and the displayComponentsWidget
        self.splitter = qt.QSplitter(qt.Qt.Vertical)
        self.splitter.addWidget(top_widget)
        self.splitter.addWidget(self._displayComponentsWidget)
        self.setCentralWidget(self.splitter)

    def setMethod(self, method: Method):
        if method in (None, ""):
            method = self.DEFAULT_METHOD
        method = Method.from_value(method)
        self.methodCB.setCurrentText(method.value)

    def getMethod(self) -> Method:
        return Method.from_value(self.methodCB.currentText())

    def setNComp(self, n_comp: int):
        if n_comp in ("0", 0):
            n_comp = self.DEFAULT_N_COMP
        self.nComponentsLE.setText(str(n_comp))

    def getNComp(self) -> int:
        return int(self.nComponentsLE.text())

    def hideButton(self):
        self._computeB.hide()

    def showButton(self):
        self._computeB.show()

    def setDataset(self, dataset: dtypes.Dataset):
        self.dataset = dataset.dataset
        self.indices = dataset.indices
        self.bg_indices = dataset.bg_indices
        self.bg_dataset = dataset.bg_dataset
        self.computeButton.setEnabled(True)
        self.detectButton.setEnabled(True)

    def _nbComponentEdited(self, *args, **kwargs):
        self.sigNbComponentsChanged.emit(int(self.nComponentsLE.text()))

    def _displayComponents(
        self, dataset: dtypes.ImageDataset, indices, bg_indices, comp, W
    ):
        """
        :param dataset: dataset for which we want to display the components
        :param indices: dataset indices
        :param bg_indices: background indices
        :param comp: components
        :param W: Matrix with the rocking curves values
        """
        if not isinstance(dataset, dtypes.ImageDataset):
            raise TypeError(
                f"dataset is expected to be an instance of {dtypes.ImageDataset}. Got {type(dataset)}."
            )
        # self._thread.finished.disconnect(self._displayComponents)
        # comp, self.W = self._thread.data
        n_comp = int(self.nComponentsLE.text())
        if comp.shape[0] < n_comp:
            n_comp = comp.shape[0]
            msg = qt.QMessageBox()
            msg.setIcon(qt.QMessageBox.Information)
            msg.setText("Found only {0} components".format(n_comp))
            msg.setStandardButtons(qt.QMessageBox.Ok)
            msg.exec()
        shape = dataset.frame_shape
        comp = comp.reshape(n_comp, shape[0], shape[1])
        self._displayComponentsWidget.show()
        self.computeButton.setEnabled(True)
        self.nComponentsLE.setEnabled(True)
        self.detectButton.setEnabled(True)
        if bg_indices is not None:
            # If filter data is activated, the matrix W has reduced dimensionality, so reshaping is not possible
            # Create empty array with shape the total number of frames
            W = numpy.zeros((dataset.nframes, n_comp))
            # Set actual values of W where threshold of filter is True
            W[indices] = W
        self._displayComponentsWidget.setComponents(
            comp,
            W,
            dataset.dims,
            dataset.get_dimensions_values(),
            dataset.title,
        )

    def _detectComp(self):
        self.detectButton.setEnabled(False)
        self.computeButton.setEnabled(False)
        self._thread = OperationThread(self, self.dataset.pca)
        self._thread.setArgs(return_vals=True)
        self._thread.finished.connect(self._setNumComp)
        self._thread.start()

    def _setNumComp(self):
        self._thread.finished.disconnect(self._setNumComp)
        vals = self._thread.data
        vals /= numpy.sum(vals)
        components = len(vals[vals > 0.01])
        self.detectButton.setEnabled(True)
        self.computeButton.setEnabled(True)
        self.nComponentsLE.setText(str(components))
        self.sigNbComponentsChanged.emit(components)

    def getDisplayComponentsWidget(self):
        return self._displayComponentsWidget



from typing import Optional

from silx.gui import qt
from silx.gui.plot.items import ItemChangedType
from silx.gui.plot.items.roi import RectangleROI


class RoiLimitsToolBar(qt.QToolBar):
    """Toolbar to edit a single rectangular ROI managed by a `RegionOfInterestManager`."""

    def __init__(self, parent=None, roiManager=None, roiIndex=0, title="ROI"):
        super().__init__(title, parent)
        assert roiManager is not None
        self._roiManager = roiManager
        self._roiManager.sigRoiChanged.connect(self._handleRoiPlotChanged)
        self._roiIndex = roiIndex
        self._initWidgets()

    @property
    def roi(self) -> RectangleROI:
        """The `RectangleROI` the toolbar is attached to."""
        rois = self._roiManager.getRois()
        return rois[self._roiIndex]

    def _initWidgets(self):
        """Create and init Toolbar widgets."""
        x0, y0 = self.roi.getOrigin()
        xs, ys = self.roi.getSize()

        self.addWidget(qt.QLabel("ROI: "))
        self.addWidget(qt.QLabel(" Origin: "))
        self._xOriginSpinBox = _createQSpinBox(self, int(x0), 0, 1000000)
        self._xOriginSpinBox.editingFinished[()].connect(self._handleRoiEditChanged)
        self.addWidget(self._xOriginSpinBox)

        self._yOriginSpinBox = _createQSpinBox(self, int(y0), 0, 1000000)
        self._yOriginSpinBox.editingFinished[()].connect(self._handleRoiEditChanged)
        self.addWidget(self._yOriginSpinBox)

        self.addWidget(qt.QLabel(" Size: "))
        self._xSizeSpinBox = _createQSpinBox(self, int(xs), 0, 1000000)
        self._xSizeSpinBox.editingFinished[()].connect(self._handleRoiEditChanged)
        self.addWidget(self._xSizeSpinBox)

        self._ySizeSpinBox = _createQSpinBox(self, int(ys), 0, 1000000)
        self._ySizeSpinBox.editingFinished[()].connect(self._handleRoiEditChanged)
        self.addWidget(self._ySizeSpinBox)

        self.roi.sigItemChanged.connect(self._handleRoiVisibiltyChanged)

    def _handleRoiPlotChanged(self) -> None:
        """Listen to ROI changes caused by the user dragging the ankor points on the plot."""
        x0, y0 = self.roi.getOrigin()
        xs, ys = self.roi.getSize()
        self._xOriginSpinBox.setValue(int(x0))
        self._yOriginSpinBox.setValue(int(y0))
        self._xSizeSpinBox.setValue(int(xs))
        self._ySizeSpinBox.setValue(int(ys))

    def _handleRoiEditChanged(self) -> None:
        """Listen to ROI changes caused by the user editing the spin boxes."""
        x0, y0 = self._xOriginSpinBox.value(), self._yOriginSpinBox.value()
        xs, ys = self._xSizeSpinBox.value(), self._ySizeSpinBox.value()
        self.roi.setGeometry(origin=(x0, y0), size=(xs, ys))

    def _handleRoiVisibiltyChanged(self, event=None) -> None:
        """Listen to ROI visibility changes caused by the user applying the ROI."""
        if event == ItemChangedType.VISIBLE:
            visible = self.roi.isVisible()
            self._xOriginSpinBox.setEnabled(visible)
            self._yOriginSpinBox.setEnabled(visible)
            self._xSizeSpinBox.setEnabled(visible)
            self._ySizeSpinBox.setEnabled(visible)


def _createQSpinBox(
    parent,
    value: Optional[int] = None,
    vmin: Optional[int] = None,
    vmax: Optional[int] = None,
) -> qt.QSpinBox:
    spinBox = qt.QSpinBox(parent=parent)
    if vmin is not None:
        spinBox.setMinimum(vmin)
    if vmax is not None:
        spinBox.setMaximum(vmax)
    if value is not None:
        spinBox.setValue(value)
    return spinBox
